"""
SEP: 0001
Title: stellar.toml
Author: stellar.org
Status: Active
Created: 2017-10-30
Updated: 2019-06-12
Version: 2.1.0
"""
from typing import Any, MutableMapping

import toml

from ..client.aiohttp_client import AiohttpClient
from ..client.base_async_client import BaseAsyncClient
from ..client.base_sync_client import BaseSyncClient
from ..client.requests_client import RequestsClient
from ..client.response import Response
from .exceptions import StellarTomlNotFoundError

__all__ = ["fetch_stellar_toml", "fetch_stellar_toml_async"]


def fetch_stellar_toml(
    domain: str,
    client: BaseSyncClient = None,
    use_http: bool = False,
) -> MutableMapping[str, Any]:
    """Retrieve the stellar.toml file from a given domain.

    Retrieve the stellar.toml file for information about interacting with
    Stellar's federation protocol for a given Stellar Anchor (specified by a
    domain).

    :param domain: The domain the .toml file is hosted at.
    :param use_http: Specifies whether the request should go over plain HTTP vs HTTPS.
        Note it is recommended that you **always** use HTTPS.
    :param client: Http Client used to send the request.
    :return: The stellar.toml file as an object via :func:`toml.loads`.
    :raises: :exc:`StellarTomlNotFoundError <stellar_sdk.sep.exceptions.StellarTomlNotFoundError>`:
        if the Stellar toml file could not be found.
    """
    if not client:
        client = RequestsClient()
    url = _build_request_url(domain, use_http)
    raw_resp = client.get(url)
    return _handle_raw_response(raw_resp)


async def fetch_stellar_toml_async(
    domain: str,
    client: BaseAsyncClient = None,
    use_http: bool = False,
) -> MutableMapping[str, Any]:
    """Retrieve the stellar.toml file from a given domain.

    Retrieve the stellar.toml file for information about interacting with
    Stellar's federation protocol for a given Stellar Anchor (specified by a
    domain).

    :param domain: The domain the .toml file is hosted at.
    :param use_http: Specifies whether the request should go over plain HTTP vs HTTPS.
        Note it is recommended that you **always** use HTTPS.
    :param client: Http Client used to send the request.
    :return: The stellar.toml file as an object via :func:`toml.loads`.
    :raises: :exc:`StellarTomlNotFoundError <stellar_sdk.sep.exceptions.StellarTomlNotFoundError>`:
        if the Stellar toml file could not be found.
    """

    if not client:
        client = AiohttpClient()
    url = _build_request_url(domain, use_http)
    raw_resp = await client.get(url)
    return _handle_raw_response(raw_resp)


def _handle_raw_response(raw_resp: Response) -> MutableMapping[str, Any]:
    if raw_resp.status_code == 404:
        raise StellarTomlNotFoundError
    resp = raw_resp.text
    return toml.loads(resp)


def _build_request_url(domain: str, use_http: bool = False) -> str:
    toml_link = "/.well-known/stellar.toml"
    protocol = "https://"
    if use_http:
        protocol = "http://"
    url = protocol + domain + toml_link
    return url



from typing import Optional, Union

from .. import xdr as stellar_xdr
from ..muxed_account import MuxedAccount
from .operation import Operation

__all__ = ["AccountMerge"]


class AccountMerge(Operation):
    """The :class:`AccountMerge` object, which represents a
    AccountMerge operation on Stellar's network.

    Transfers the native balance (the amount of XLM an account holds) to
    another account and removes the source account from the ledger.

    Threshold: High

    See `Account Merge <https://developers.stellar.org/docs/start/list-of-operations/#account-merge>`_ for more information.

    :param destination: Destination to merge the source account into.
    :param source: The source account for the operation. Defaults to the transaction's source account.

    """

    _XDR_OPERATION_TYPE: stellar_xdr.OperationType = (
        stellar_xdr.OperationType.ACCOUNT_MERGE
    )

    def __init__(
        self,
        destination: Union[MuxedAccount, str],
        source: Optional[Union[MuxedAccount, str]] = None,
    ) -> None:
        super().__init__(source)
        if isinstance(destination, str):
            destination = MuxedAccount.from_account(destination)
        self.destination: MuxedAccount = destination

    def _to_operation_body(self) -> stellar_xdr.OperationBody:
        destination = self.destination.to_xdr_object()
        body = stellar_xdr.OperationBody(
            type=self._XDR_OPERATION_TYPE, destination=destination
        )
        return body

    @classmethod
    def from_xdr_object(cls, xdr_object: stellar_xdr.Operation) -> "AccountMerge":
        """Creates a :class:`AccountMerge` object from an XDR Operation
        object.

        """
        source = Operation.get_source_from_xdr_obj(xdr_object)
        assert xdr_object.body.destination is not None
        destination = MuxedAccount.from_xdr_object(xdr_object.body.destination)
        op = cls(source=source, destination=destination)
        return op

    def __str__(self):
        return f"<AccountMerge [destination={self.destination}, source={self.source}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64

from xdrlib3 import Packer, Unpacker

from .int64 import Int64
from .pool_id import PoolID

__all__ = ["LiquidityPoolWithdrawOp"]


class LiquidityPoolWithdrawOp:
    """
    XDR Source Code::

        struct LiquidityPoolWithdrawOp
        {
            PoolID liquidityPoolID;
            int64 amount;     // amount of pool shares to withdraw
            int64 minAmountA; // minimum amount of first asset to withdraw
            int64 minAmountB; // minimum amount of second asset to withdraw
        };
    """

    def __init__(
        self,
        liquidity_pool_id: PoolID,
        amount: Int64,
        min_amount_a: Int64,
        min_amount_b: Int64,
    ) -> None:
        self.liquidity_pool_id = liquidity_pool_id
        self.amount = amount
        self.min_amount_a = min_amount_a
        self.min_amount_b = min_amount_b

    def pack(self, packer: Packer) -> None:
        self.liquidity_pool_id.pack(packer)
        self.amount.pack(packer)
        self.min_amount_a.pack(packer)
        self.min_amount_b.pack(packer)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> LiquidityPoolWithdrawOp:
        liquidity_pool_id = PoolID.unpack(unpacker)
        amount = Int64.unpack(unpacker)
        min_amount_a = Int64.unpack(unpacker)
        min_amount_b = Int64.unpack(unpacker)
        return cls(
            liquidity_pool_id=liquidity_pool_id,
            amount=amount,
            min_amount_a=min_amount_a,
            min_amount_b=min_amount_b,
        )

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> LiquidityPoolWithdrawOp:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> LiquidityPoolWithdrawOp:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.liquidity_pool_id,
                self.amount,
                self.min_amount_a,
                self.min_amount_b,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.liquidity_pool_id == other.liquidity_pool_id
            and self.amount == other.amount
            and self.min_amount_a == other.min_amount_a
            and self.min_amount_b == other.min_amount_b
        )

    def __str__(self):
        out = [
            f"liquidity_pool_id={self.liquidity_pool_id}",
            f"amount={self.amount}",
            f"min_amount_a={self.min_amount_a}",
            f"min_amount_b={self.min_amount_b}",
        ]
        return f"<LiquidityPoolWithdrawOp [{', '.join(out)}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64

from xdrlib3 import Packer, Unpacker

from .extension_point import ExtensionPoint
from .int64 import Int64
from .soroban_resources import SorobanResources

__all__ = ["SorobanTransactionData"]


class SorobanTransactionData:
    """
    XDR Source Code::

        struct SorobanTransactionData
        {
            ExtensionPoint ext;
            SorobanResources resources;
            // Amount of the transaction `fee` allocated to the Soroban resource fees.
            // The fraction of `resourceFee` corresponding to `resources` specified
            // above is *not* refundable (i.e. fees for instructions, ledger I/O), as
            // well as fees for the transaction size.
            // The remaining part of the fee is refundable and the charged value is
            // based on the actual consumption of refundable resources (events, ledger
            // rent bumps).
            // The `inclusionFee` used for prioritization of the transaction is defined
            // as `tx.fee - resourceFee`.
            int64 resourceFee;
        };
    """

    def __init__(
        self,
        ext: ExtensionPoint,
        resources: SorobanResources,
        resource_fee: Int64,
    ) -> None:
        self.ext = ext
        self.resources = resources
        self.resource_fee = resource_fee

    def pack(self, packer: Packer) -> None:
        self.ext.pack(packer)
        self.resources.pack(packer)
        self.resource_fee.pack(packer)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> SorobanTransactionData:
        ext = ExtensionPoint.unpack(unpacker)
        resources = SorobanResources.unpack(unpacker)
        resource_fee = Int64.unpack(unpacker)
        return cls(
            ext=ext,
            resources=resources,
            resource_fee=resource_fee,
        )

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> SorobanTransactionData:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> SorobanTransactionData:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.ext,
                self.resources,
                self.resource_fee,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.ext == other.ext
            and self.resources == other.resources
            and self.resource_fee == other.resource_fee
        )

    def __str__(self):
        out = [
            f"ext={self.ext}",
            f"resources={self.resources}",
            f"resource_fee={self.resource_fee}",
        ]
        return f"<SorobanTransactionData [{', '.join(out)}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64

from xdrlib3 import Packer, Unpacker

from .ledger_entry import LedgerEntry
from .ledger_entry_change_type import LedgerEntryChangeType
from .ledger_key import LedgerKey

__all__ = ["LedgerEntryChange"]


class LedgerEntryChange:
    """
    XDR Source Code::

        union LedgerEntryChange switch (LedgerEntryChangeType type)
        {
        case LEDGER_ENTRY_CREATED:
            LedgerEntry created;
        case LEDGER_ENTRY_UPDATED:
            LedgerEntry updated;
        case LEDGER_ENTRY_REMOVED:
            LedgerKey removed;
        case LEDGER_ENTRY_STATE:
            LedgerEntry state;
        };
    """

    def __init__(
        self,
        type: LedgerEntryChangeType,
        created: LedgerEntry = None,
        updated: LedgerEntry = None,
        removed: LedgerKey = None,
        state: LedgerEntry = None,
    ) -> None:
        self.type = type
        self.created = created
        self.updated = updated
        self.removed = removed
        self.state = state

    def pack(self, packer: Packer) -> None:
        self.type.pack(packer)
        if self.type == LedgerEntryChangeType.LEDGER_ENTRY_CREATED:
            if self.created is None:
                raise ValueError("created should not be None.")
            self.created.pack(packer)
            return
        if self.type == LedgerEntryChangeType.LEDGER_ENTRY_UPDATED:
            if self.updated is None:
                raise ValueError("updated should not be None.")
            self.updated.pack(packer)
            return
        if self.type == LedgerEntryChangeType.LEDGER_ENTRY_REMOVED:
            if self.removed is None:
                raise ValueError("removed should not be None.")
            self.removed.pack(packer)
            return
        if self.type == LedgerEntryChangeType.LEDGER_ENTRY_STATE:
            if self.state is None:
                raise ValueError("state should not be None.")
            self.state.pack(packer)
            return

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> LedgerEntryChange:
        type = LedgerEntryChangeType.unpack(unpacker)
        if type == LedgerEntryChangeType.LEDGER_ENTRY_CREATED:
            created = LedgerEntry.unpack(unpacker)
            return cls(type=type, created=created)
        if type == LedgerEntryChangeType.LEDGER_ENTRY_UPDATED:
            updated = LedgerEntry.unpack(unpacker)
            return cls(type=type, updated=updated)
        if type == LedgerEntryChangeType.LEDGER_ENTRY_REMOVED:
            removed = LedgerKey.unpack(unpacker)
            return cls(type=type, removed=removed)
        if type == LedgerEntryChangeType.LEDGER_ENTRY_STATE:
            state = LedgerEntry.unpack(unpacker)
            return cls(type=type, state=state)
        return cls(type=type)

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> LedgerEntryChange:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> LedgerEntryChange:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.type,
                self.created,
                self.updated,
                self.removed,
                self.state,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.type == other.type
            and self.created == other.created
            and self.updated == other.updated
            and self.removed == other.removed
            and self.state == other.state
        )

    def __str__(self):
        out = []
        out.append(f"type={self.type}")
        out.append(f"created={self.created}") if self.created is not None else None
        out.append(f"updated={self.updated}") if self.updated is not None else None
        out.append(f"removed={self.removed}") if self.removed is not None else None
        out.append(f"state={self.state}") if self.state is not None else None
        return f"<LedgerEntryChange [{', '.join(out)}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64
from typing import List

from xdrlib3 import Packer, Unpacker

from .inflation_payout import InflationPayout
from .inflation_result_code import InflationResultCode

__all__ = ["InflationResult"]


class InflationResult:
    """
    XDR Source Code::

        union InflationResult switch (InflationResultCode code)
        {
        case INFLATION_SUCCESS:
            InflationPayout payouts<>;
        case INFLATION_NOT_TIME:
            void;
        };
    """

    def __init__(
        self,
        code: InflationResultCode,
        payouts: List[InflationPayout] = None,
    ) -> None:
        _expect_max_length = 4294967295
        if payouts and len(payouts) > _expect_max_length:
            raise ValueError(
                f"The maximum length of `payouts` should be {_expect_max_length}, but got {len(payouts)}."
            )
        self.code = code
        self.payouts = payouts

    def pack(self, packer: Packer) -> None:
        self.code.pack(packer)
        if self.code == InflationResultCode.INFLATION_SUCCESS:
            if self.payouts is None:
                raise ValueError("payouts should not be None.")
            packer.pack_uint(len(self.payouts))
            for payouts_item in self.payouts:
                payouts_item.pack(packer)
            return
        if self.code == InflationResultCode.INFLATION_NOT_TIME:
            return

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> InflationResult:
        code = InflationResultCode.unpack(unpacker)
        if code == InflationResultCode.INFLATION_SUCCESS:
            length = unpacker.unpack_uint()
            payouts = []
            for _ in range(length):
                payouts.append(InflationPayout.unpack(unpacker))
            return cls(code=code, payouts=payouts)
        if code == InflationResultCode.INFLATION_NOT_TIME:
            return cls(code=code)
        return cls(code=code)

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> InflationResult:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> InflationResult:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.code,
                self.payouts,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.code == other.code and self.payouts == other.payouts

    def __str__(self):
        out = []
        out.append(f"code={self.code}")
        out.append(f"payouts={self.payouts}") if self.payouts is not None else None
        return f"<InflationResult [{', '.join(out)}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64
from typing import List, Optional

from xdrlib3 import Packer, Unpacker

from .account_entry_ext import AccountEntryExt
from .account_id import AccountID
from .constants import *
from .int64 import Int64
from .sequence_number import SequenceNumber
from .signer import Signer
from .string32 import String32
from .thresholds import Thresholds
from .uint32 import Uint32

__all__ = ["AccountEntry"]


class AccountEntry:
    """
    XDR Source Code::

        struct AccountEntry
        {
            AccountID accountID;      // master public key for this account
            int64 balance;            // in stroops
            SequenceNumber seqNum;    // last sequence number used for this account
            uint32 numSubEntries;     // number of sub-entries this account has
                                      // drives the reserve
            AccountID* inflationDest; // Account to vote for during inflation
            uint32 flags;             // see AccountFlags

            string32 homeDomain; // can be used for reverse federation and memo lookup

            // fields used for signatures
            // thresholds stores unsigned bytes: [weight of master|low|medium|high]
            Thresholds thresholds;

            Signer signers<MAX_SIGNERS>; // possible signers for this account

            // reserved for future use
            union switch (int v)
            {
            case 0:
                void;
            case 1:
                AccountEntryExtensionV1 v1;
            }
            ext;
        };
    """

    def __init__(
        self,
        account_id: AccountID,
        balance: Int64,
        seq_num: SequenceNumber,
        num_sub_entries: Uint32,
        inflation_dest: Optional[AccountID],
        flags: Uint32,
        home_domain: String32,
        thresholds: Thresholds,
        signers: List[Signer],
        ext: AccountEntryExt,
    ) -> None:
        _expect_max_length = MAX_SIGNERS
        if signers and len(signers) > _expect_max_length:
            raise ValueError(
                f"The maximum length of `signers` should be {_expect_max_length}, but got {len(signers)}."
            )
        self.account_id = account_id
        self.balance = balance
        self.seq_num = seq_num
        self.num_sub_entries = num_sub_entries
        self.inflation_dest = inflation_dest
        self.flags = flags
        self.home_domain = home_domain
        self.thresholds = thresholds
        self.signers = signers
        self.ext = ext

    def pack(self, packer: Packer) -> None:
        self.account_id.pack(packer)
        self.balance.pack(packer)
        self.seq_num.pack(packer)
        self.num_sub_entries.pack(packer)
        if self.inflation_dest is None:
            packer.pack_uint(0)
        else:
            packer.pack_uint(1)
            self.inflation_dest.pack(packer)
        self.flags.pack(packer)
        self.home_domain.pack(packer)
        self.thresholds.pack(packer)
        packer.pack_uint(len(self.signers))
        for signers_item in self.signers:
            signers_item.pack(packer)
        self.ext.pack(packer)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> AccountEntry:
        account_id = AccountID.unpack(unpacker)
        balance = Int64.unpack(unpacker)
        seq_num = SequenceNumber.unpack(unpacker)
        num_sub_entries = Uint32.unpack(unpacker)
        inflation_dest = AccountID.unpack(unpacker) if unpacker.unpack_uint() else None
        flags = Uint32.unpack(unpacker)
        home_domain = String32.unpack(unpacker)
        thresholds = Thresholds.unpack(unpacker)
        length = unpacker.unpack_uint()
        signers = []
        for _ in range(length):
            signers.append(Signer.unpack(unpacker))
        ext = AccountEntryExt.unpack(unpacker)
        return cls(
            account_id=account_id,
            balance=balance,
            seq_num=seq_num,
            num_sub_entries=num_sub_entries,
            inflation_dest=inflation_dest,
            flags=flags,
            home_domain=home_domain,
            thresholds=thresholds,
            signers=signers,
            ext=ext,
        )

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> AccountEntry:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> AccountEntry:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.account_id,
                self.balance,
                self.seq_num,
                self.num_sub_entries,
                self.inflation_dest,
                self.flags,
                self.home_domain,
                self.thresholds,
                self.signers,
                self.ext,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.account_id == other.account_id
            and self.balance == other.balance
            and self.seq_num == other.seq_num
            and self.num_sub_entries == other.num_sub_entries
            and self.inflation_dest == other.inflation_dest
            and self.flags == other.flags
            and self.home_domain == other.home_domain
            and self.thresholds == other.thresholds
            and self.signers == other.signers
            and self.ext == other.ext
        )

    def __str__(self):
        out = [
            f"account_id={self.account_id}",
            f"balance={self.balance}",
            f"seq_num={self.seq_num}",
            f"num_sub_entries={self.num_sub_entries}",
            f"inflation_dest={self.inflation_dest}",
            f"flags={self.flags}",
            f"home_domain={self.home_domain}",
            f"thresholds={self.thresholds}",
            f"signers={self.signers}",
            f"ext={self.ext}",
        ]
        return f"<AccountEntry [{', '.join(out)}]>"



from decimal import Decimal
from typing import Optional, Union

from .. import xdr as stellar_xdr
from ..asset import Asset
from ..liquidity_pool_asset import LiquidityPoolAsset
from ..muxed_account import MuxedAccount
from ..utils import raise_if_not_valid_amount
from .operation import Operation

__all__ = ["ChangeTrust"]


class ChangeTrust(Operation):
    """The :class:`ChangeTrust` object, which represents a ChangeTrust
    operation on Stellar's network.

    Creates, updates, or deletes a trustline. For more on trustlines, please
    refer to the `assets documentation
    <https://developers.stellar.org/docs/glossary/assets/>`_.

    Threshold: Medium

    See `Change Trust <https://developers.stellar.org/docs/start/list-of-operations/#change-trust>`_ for more information.

    :param asset: The asset for the trust line.
    :param limit: The limit for the asset, defaults to max int64(``922337203685.4775807``).
        If the limit is set to ``"0"`` it deletes the trustline.
    :param source: The source account for the operation. Defaults to the transaction's source account.

    """

    _DEFAULT_LIMIT = "922337203685.4775807"

    _XDR_OPERATION_TYPE: stellar_xdr.OperationType = (
        stellar_xdr.OperationType.CHANGE_TRUST
    )

    def __init__(
        self,
        asset: Union[Asset, LiquidityPoolAsset],
        limit: Union[str, Decimal, None] = None,  # _DEFAULT_LIMIT default
        source: Optional[Union[MuxedAccount, str]] = None,
    ) -> None:
        super().__init__(source)
        self.asset = asset
        if limit is None:
            self.limit: str = self._DEFAULT_LIMIT
        else:
            limit = str(limit)
            self.limit = limit
        raise_if_not_valid_amount(self.limit, "limit")

    def _to_operation_body(self) -> stellar_xdr.OperationBody:
        line = self.asset.to_change_trust_asset_xdr_object()
        limit = stellar_xdr.Int64(Operation.to_xdr_amount(self.limit))
        change_trust_op = stellar_xdr.ChangeTrustOp(line, limit)
        body = stellar_xdr.OperationBody(
            type=self._XDR_OPERATION_TYPE, change_trust_op=change_trust_op
        )
        return body

    @classmethod
    def from_xdr_object(cls, xdr_object: stellar_xdr.Operation) -> "ChangeTrust":
        """Creates a :class:`ChangeTrust` object from an XDR Operation
        object.

        """
        source = Operation.get_source_from_xdr_obj(xdr_object)
        assert xdr_object.body.change_trust_op is not None
        if (
            xdr_object.body.change_trust_op.line.type
            == stellar_xdr.AssetType.ASSET_TYPE_POOL_SHARE
        ):
            line: Union[Asset, LiquidityPoolAsset] = LiquidityPoolAsset.from_xdr_object(
                xdr_object.body.change_trust_op.line
            )
        else:
            line = Asset.from_xdr_object(xdr_object.body.change_trust_op.line)
        limit = Operation.from_xdr_amount(xdr_object.body.change_trust_op.limit.int64)
        op = cls(source=source, asset=line, limit=limit)
        return op

    def __str__(self):
        return f"<ChangeTrust [asset={self.asset}, limit={self.limit}, source={self.source}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64

from xdrlib3 import Packer, Unpacker

from .config_upgrade_set_key import ConfigUpgradeSetKey
from .ledger_upgrade_type import LedgerUpgradeType
from .uint32 import Uint32

__all__ = ["LedgerUpgrade"]


class LedgerUpgrade:
    """
    XDR Source Code::

        union LedgerUpgrade switch (LedgerUpgradeType type)
        {
        case LEDGER_UPGRADE_VERSION:
            uint32 newLedgerVersion; // update ledgerVersion
        case LEDGER_UPGRADE_BASE_FEE:
            uint32 newBaseFee; // update baseFee
        case LEDGER_UPGRADE_MAX_TX_SET_SIZE:
            uint32 newMaxTxSetSize; // update maxTxSetSize
        case LEDGER_UPGRADE_BASE_RESERVE:
            uint32 newBaseReserve; // update baseReserve
        case LEDGER_UPGRADE_FLAGS:
            uint32 newFlags; // update flags
        case LEDGER_UPGRADE_CONFIG:
            // Update arbitrary `ConfigSetting` entries identified by the key.
            ConfigUpgradeSetKey newConfig;
        case LEDGER_UPGRADE_MAX_SOROBAN_TX_SET_SIZE:
            // Update ConfigSettingContractExecutionLanesV0.ledgerMaxTxCount without
            // using `LEDGER_UPGRADE_CONFIG`.
            uint32 newMaxSorobanTxSetSize;
        };
    """

    def __init__(
        self,
        type: LedgerUpgradeType,
        new_ledger_version: Uint32 = None,
        new_base_fee: Uint32 = None,
        new_max_tx_set_size: Uint32 = None,
        new_base_reserve: Uint32 = None,
        new_flags: Uint32 = None,
        new_config: ConfigUpgradeSetKey = None,
        new_max_soroban_tx_set_size: Uint32 = None,
    ) -> None:
        self.type = type
        self.new_ledger_version = new_ledger_version
        self.new_base_fee = new_base_fee
        self.new_max_tx_set_size = new_max_tx_set_size
        self.new_base_reserve = new_base_reserve
        self.new_flags = new_flags
        self.new_config = new_config
        self.new_max_soroban_tx_set_size = new_max_soroban_tx_set_size

    def pack(self, packer: Packer) -> None:
        self.type.pack(packer)
        if self.type == LedgerUpgradeType.LEDGER_UPGRADE_VERSION:
            if self.new_ledger_version is None:
                raise ValueError("new_ledger_version should not be None.")
            self.new_ledger_version.pack(packer)
            return
        if self.type == LedgerUpgradeType.LEDGER_UPGRADE_BASE_FEE:
            if self.new_base_fee is None:
                raise ValueError("new_base_fee should not be None.")
            self.new_base_fee.pack(packer)
            return
        if self.type == LedgerUpgradeType.LEDGER_UPGRADE_MAX_TX_SET_SIZE:
            if self.new_max_tx_set_size is None:
                raise ValueError("new_max_tx_set_size should not be None.")
            self.new_max_tx_set_size.pack(packer)
            return
        if self.type == LedgerUpgradeType.LEDGER_UPGRADE_BASE_RESERVE:
            if self.new_base_reserve is None:
                raise ValueError("new_base_reserve should not be None.")
            self.new_base_reserve.pack(packer)
            return
        if self.type == LedgerUpgradeType.LEDGER_UPGRADE_FLAGS:
            if self.new_flags is None:
                raise ValueError("new_flags should not be None.")
            self.new_flags.pack(packer)
            return
        if self.type == LedgerUpgradeType.LEDGER_UPGRADE_CONFIG:
            if self.new_config is None:
                raise ValueError("new_config should not be None.")
            self.new_config.pack(packer)
            return
        if self.type == LedgerUpgradeType.LEDGER_UPGRADE_MAX_SOROBAN_TX_SET_SIZE:
            if self.new_max_soroban_tx_set_size is None:
                raise ValueError("new_max_soroban_tx_set_size should not be None.")
            self.new_max_soroban_tx_set_size.pack(packer)
            return

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> LedgerUpgrade:
        type = LedgerUpgradeType.unpack(unpacker)
        if type == LedgerUpgradeType.LEDGER_UPGRADE_VERSION:
            new_ledger_version = Uint32.unpack(unpacker)
            return cls(type=type, new_ledger_version=new_ledger_version)
        if type == LedgerUpgradeType.LEDGER_UPGRADE_BASE_FEE:
            new_base_fee = Uint32.unpack(unpacker)
            return cls(type=type, new_base_fee=new_base_fee)
        if type == LedgerUpgradeType.LEDGER_UPGRADE_MAX_TX_SET_SIZE:
            new_max_tx_set_size = Uint32.unpack(unpacker)
            return cls(type=type, new_max_tx_set_size=new_max_tx_set_size)
        if type == LedgerUpgradeType.LEDGER_UPGRADE_BASE_RESERVE:
            new_base_reserve = Uint32.unpack(unpacker)
            return cls(type=type, new_base_reserve=new_base_reserve)
        if type == LedgerUpgradeType.LEDGER_UPGRADE_FLAGS:
            new_flags = Uint32.unpack(unpacker)
            return cls(type=type, new_flags=new_flags)
        if type == LedgerUpgradeType.LEDGER_UPGRADE_CONFIG:
            new_config = ConfigUpgradeSetKey.unpack(unpacker)
            return cls(type=type, new_config=new_config)
        if type == LedgerUpgradeType.LEDGER_UPGRADE_MAX_SOROBAN_TX_SET_SIZE:
            new_max_soroban_tx_set_size = Uint32.unpack(unpacker)
            return cls(
                type=type, new_max_soroban_tx_set_size=new_max_soroban_tx_set_size
            )
        return cls(type=type)

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> LedgerUpgrade:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> LedgerUpgrade:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.type,
                self.new_ledger_version,
                self.new_base_fee,
                self.new_max_tx_set_size,
                self.new_base_reserve,
                self.new_flags,
                self.new_config,
                self.new_max_soroban_tx_set_size,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.type == other.type
            and self.new_ledger_version == other.new_ledger_version
            and self.new_base_fee == other.new_base_fee
            and self.new_max_tx_set_size == other.new_max_tx_set_size
            and self.new_base_reserve == other.new_base_reserve
            and self.new_flags == other.new_flags
            and self.new_config == other.new_config
            and self.new_max_soroban_tx_set_size == other.new_max_soroban_tx_set_size
        )

    def __str__(self):
        out = []
        out.append(f"type={self.type}")
        out.append(
            f"new_ledger_version={self.new_ledger_version}"
        ) if self.new_ledger_version is not None else None
        out.append(
            f"new_base_fee={self.new_base_fee}"
        ) if self.new_base_fee is not None else None
        out.append(
            f"new_max_tx_set_size={self.new_max_tx_set_size}"
        ) if self.new_max_tx_set_size is not None else None
        out.append(
            f"new_base_reserve={self.new_base_reserve}"
        ) if self.new_base_reserve is not None else None
        out.append(
            f"new_flags={self.new_flags}"
        ) if self.new_flags is not None else None
        out.append(
            f"new_config={self.new_config}"
        ) if self.new_config is not None else None
        out.append(
            f"new_max_soroban_tx_set_size={self.new_max_soroban_tx_set_size}"
        ) if self.new_max_soroban_tx_set_size is not None else None
        return f"<LedgerUpgrade [{', '.join(out)}]>"



import binascii

from . import xdr as stellar_xdr
from .utils import raise_if_not_valid_hash

__all__ = ["LiquidityPoolId"]


class LiquidityPoolId:
    """The :class:`LiquidityPoolId` object, which represents the asset referenced by a trustline to a liquidity pool.

    :param liquidity_pool_id: The ID of the liquidity pool in hex string.
    :raise: :exc:`ValueError <stellar_sdk.exceptions.ValueError>`
    """

    def __init__(self, liquidity_pool_id: str) -> None:
        self.liquidity_pool_id: str = liquidity_pool_id
        self.type: str = "liquidity_pool_shares"
        raise_if_not_valid_hash(self.liquidity_pool_id, "liquidity_pool_id")

    def to_trust_line_asset_xdr_object(self) -> stellar_xdr.TrustLineAsset:
        """Returns the xdr object for this LiquidityPoolId object.

        :return: XDR TrustLineAsset object
        """
        liquidity_pool_id_bytes = binascii.unhexlify(self.liquidity_pool_id)
        liquidity_pool_id = stellar_xdr.PoolID.from_xdr_bytes(liquidity_pool_id_bytes)
        return stellar_xdr.TrustLineAsset(
            stellar_xdr.AssetType.ASSET_TYPE_POOL_SHARE,
            liquidity_pool_id=liquidity_pool_id,
        )

    @classmethod
    def from_xdr_object(
        cls, xdr_object: stellar_xdr.TrustLineAsset
    ) -> "LiquidityPoolId":
        """Create a :class:`LiquidityPoolId` from an XDR Asset object.

        :param xdr_object: The XDR TrustLineAsset object.
        :return: A new :class:`LiquidityPoolId` object from the given XDR TrustLineAsset object.
        """
        if xdr_object.type != stellar_xdr.AssetType.ASSET_TYPE_POOL_SHARE:
            raise ValueError(f"Unexpected asset type: {xdr_object.type}")
        assert xdr_object.liquidity_pool_id is not None
        liquidity_pool_id = xdr_object.liquidity_pool_id.pool_id.hash.hex()
        return cls(liquidity_pool_id)

    def __hash__(self):
        return hash((self.liquidity_pool_id, self.type))

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.liquidity_pool_id == other.liquidity_pool_id

    def __str__(self):
        return f"<LiquidityPoolId [liquidity_pool_id={self.liquidity_pool_id}]>"



import base64
import binascii
import struct
from enum import Enum

from xdrlib3 import Packer, Unpacker

from . import xdr as stellar_xdr
from .exceptions import (
    Ed25519PublicKeyInvalidError,
    Ed25519SecretSeedInvalidError,
    MuxedEd25519AccountInvalidError,
)

__all__ = ["StrKey"]


class _VersionByte(Enum):
    ED25519_PUBLIC_KEY = binascii.a2b_hex("30")  # G 48 6 << 3
    ED25519_SECRET_SEED = binascii.a2b_hex("90")  # S 144 18 << 3
    PRE_AUTH_TX = binascii.a2b_hex("98")  # T 152 19 << 3
    SHA256_HASH = binascii.a2b_hex("b8")  # X 184 23 << 3
    MUXED_ACCOUNT = binascii.a2b_hex("60")  # M 96 12 << 3
    ED25519_SIGNED_PAYLOAD = binascii.a2b_hex("78")  # P 120 15 << 3
    CONTRACT = binascii.a2b_hex("10")  # C 16 2 << 3


class StrKey:
    """StrKey is a helper class that allows encoding and decoding strkey."""

    @staticmethod
    def encode_ed25519_public_key(data: bytes) -> str:
        """Encodes data to encoded ed25519 public key strkey.

        :param data: data to encode
        :return: encoded ed25519 public key strkey
        :raises:
            :exc:`ValueError <stellar_sdk.exceptions.ValueError>`
        """
        return _encode_check(_VersionByte.ED25519_PUBLIC_KEY, data)

    @staticmethod
    def decode_ed25519_public_key(data: str) -> bytes:
        """Decodes encoded ed25519 public key strkey to raw data.

        :param data: encoded ed25519 public key strkey
        :return: raw bytes
        :raises:
            :exc:`Ed25519PublicKeyInvalidError <stellar_sdk.exceptions.Ed25519PublicKeyInvalidError>`
        """
        try:
            return _decode_check(_VersionByte.ED25519_PUBLIC_KEY, data)
        except Exception:
            raise Ed25519PublicKeyInvalidError(f"Invalid Ed25519 Public Key: {data}")

    @staticmethod
    def is_valid_ed25519_public_key(public_key: str) -> bool:
        """Returns ``True`` if the given `seed` is a valid ed25519 public key strkey.

        :param public_key: encoded ed25519 public key strkey
        :return: ``True`` if the given key is valid
        """
        return _is_valid(_VersionByte.ED25519_PUBLIC_KEY, public_key)

    @staticmethod
    def encode_ed25519_secret_seed(data: bytes) -> str:
        """Encodes data to encoded ed25519 secret seed strkey.

        :param data: data to encode
        :return: encoded ed25519 secret seed strkey
        :raises:
            :exc:`ValueError <stellar_sdk.exceptions.ValueError>`
        """
        return _encode_check(_VersionByte.ED25519_SECRET_SEED, data)

    @staticmethod
    def decode_ed25519_secret_seed(data: str) -> bytes:
        """Decodes encoded ed25519 secret seed strkey to raw data.

        :param data: encoded ed25519 secret seed strkey
        :return: raw bytes
        :raises:
            :exc:`Ed25519SecretSeedInvalidError <stellar_sdk.exceptions.Ed25519SecretSeedInvalidError>`
        """
        try:
            return _decode_check(_VersionByte.ED25519_SECRET_SEED, data)
        except Exception:
            raise Ed25519SecretSeedInvalidError(f"Invalid Ed25519 Secret Seed: {data}")

    @staticmethod
    def is_valid_ed25519_secret_seed(seed: str) -> bool:
        """Returns ``True`` if the given `seed` is a valid ed25519 secret seed strkey.

        :param seed: encoded ed25519 secret seed strkey
        :return: ``True`` if the given key is valid
        """
        return _is_valid(_VersionByte.ED25519_SECRET_SEED, seed)

    @staticmethod
    def encode_pre_auth_tx(data: bytes) -> str:
        """Encodes data to encoded pre auth tx strkey.

        :param data: data to encode
        :return: encoded pre auth tx strkey
        :raises:
            :exc:`ValueError <stellar_sdk.exceptions.ValueError>`
        """
        return _encode_check(_VersionByte.PRE_AUTH_TX, data)

    @staticmethod
    def decode_pre_auth_tx(data: str) -> bytes:
        """Decodes encoded pre auth tx strkey to raw data.

        :param data: encoded pre auth tx strkey
        :return: raw bytes
        :raises:
            :exc:`ValueError <stellar_sdk.exceptions.ValueError>`
        """
        try:
            return _decode_check(_VersionByte.PRE_AUTH_TX, data)
        except Exception as e:
            raise ValueError(f"Invalid Pre Auth Tx Key: {data}") from e

    @staticmethod
    def is_valid_pre_auth_tx(pre_auth_tx: str) -> bool:
        """Returns ``True`` if the given `pre_auth_tx` is a valid encoded pre auth tx strkey.

        :param pre_auth_tx: encoded pre auth tx strkey
        :return: ``True`` if the given key is valid
        """
        return _is_valid(_VersionByte.PRE_AUTH_TX, pre_auth_tx)

    @staticmethod
    def encode_sha256_hash(data: bytes) -> str:
        """Encodes data to encoded sha256 hash strkey.

        :param data: data to encode
        :return: encoded sha256 hash strkey
        :raises:
            :exc:`ValueError <stellar_sdk.exceptions.ValueError>`
        """
        return _encode_check(_VersionByte.SHA256_HASH, data)

    @staticmethod
    def decode_sha256_hash(data: str) -> bytes:
        """Decodes encoded sha256 hash strkey to raw data.

        :param data: encoded sha256 hash strkey
        :return: raw bytes
        :raises:
            :exc:`ValueError <stellar_sdk.exceptions.ValueError>`
        """
        try:
            return _decode_check(_VersionByte.SHA256_HASH, data)
        except Exception as e:
            raise ValueError(f"Invalid sha256 Hash Key: {data}") from e

    @staticmethod
    def is_valid_sha256_hash(sha256_hash: str) -> bool:
        """Returns ``True`` if the given `sha256_hash` is a valid encoded sha256 hash(HashX) strkey.

        :param sha256_hash: encoded sha256 hash(HashX) strkey
        :return: ``True`` if the given key is valid
        """
        return _is_valid(_VersionByte.SHA256_HASH, sha256_hash)

    @staticmethod
    def encode_muxed_account(data: stellar_xdr.MuxedAccount) -> str:
        """Encodes data to encoded muxed account strkey.

        :param data: data to encode
        :return: encoded muxed account strkey
        :raises:
            :exc:`ValueError <stellar_sdk.exceptions.ValueError>`
        """
        if data.type == stellar_xdr.CryptoKeyType.KEY_TYPE_ED25519:
            assert data.ed25519 is not None
            return StrKey.encode_ed25519_public_key(data.ed25519.uint256)

        assert data.med25519 is not None
        packer = Packer()
        data.med25519.ed25519.pack(packer)
        data.med25519.id.pack(packer)
        return _encode_check(_VersionByte.MUXED_ACCOUNT, packer.get_buffer())

    @staticmethod
    def decode_muxed_account(data: str) -> stellar_xdr.MuxedAccount:
        """Decodes encoded muxed account strkey to raw data.

        :param data: encoded muxed account strkey
        :return: raw bytes
        :raises:
            :exc:`ValueError <stellar_sdk.exceptions.ValueError>`
        """
        data_length = len(data)
        if data_length == 56:
            muxed = stellar_xdr.MuxedAccount(
                type=stellar_xdr.CryptoKeyType.KEY_TYPE_ED25519,
                ed25519=stellar_xdr.Uint256(StrKey.decode_ed25519_public_key(data)),
            )
        elif data_length == 69:
            # let's optimize it in v3.
            try:
                xdr_bytes = _decode_check(_VersionByte.MUXED_ACCOUNT, data)
            except Exception:
                raise MuxedEd25519AccountInvalidError(
                    "Invalid Muxed Account: {}".format(data)
                )

            unpacker = Unpacker(xdr_bytes)
            ed25519 = stellar_xdr.Uint256.unpack(unpacker)
            id = stellar_xdr.Uint64.unpack(unpacker)
            med25519 = stellar_xdr.MuxedAccountMed25519(
                id=id,
                ed25519=ed25519,
            )
            muxed = stellar_xdr.MuxedAccount(
                type=stellar_xdr.CryptoKeyType.KEY_TYPE_MUXED_ED25519, med25519=med25519
            )
        else:
            raise ValueError("Invalid encoded string, this is not a valid account.")
        return muxed

    @staticmethod
    def encode_ed25519_signed_payload(data: bytes) -> str:
        """Encodes data to encoded ed25519 signed payload strkey.

        :param data: data to encode
        :return: encoded ed25519 signed payload strkey
        :raises:
            :exc:`ValueError <stellar_sdk.exceptions.ValueError>`
        """
        return _encode_check(_VersionByte.ED25519_SIGNED_PAYLOAD, data)

    @staticmethod
    def decode_ed25519_signed_payload(data: str) -> bytes:
        """Decodes encoded ed25519 signed payload strkey to raw data.

        :param data: encoded ed25519 signed payload strkey
        :return: raw bytes
        :raises:
            :exc:`ValueError <stellar_sdk.exceptions.ValueError>`
        """
        try:
            return _decode_check(_VersionByte.ED25519_SIGNED_PAYLOAD, data)
        except Exception as e:
            raise ValueError(f"Invalid Ed25519 Signed Payload Key: {data}") from e

    @staticmethod
    def is_valid_ed25519_signed_payload(ed25519_signed_payload: str) -> bool:
        """Returns ``True`` if the given `ed25519_signed_payload` is a valid encoded ed25519 signed payload strkey.

        :param ed25519_signed_payload: encoded ed25519 signed payload strkey
        :return: ``True`` if the given key is valid
        """
        return _is_valid(_VersionByte.ED25519_SIGNED_PAYLOAD, ed25519_signed_payload)

    @staticmethod
    def encode_contract(data: bytes) -> str:
        """Encodes data to encoded contract strkey.

        :param data: data to encode
        :return: encoded contract strkey
        :raises:
            :exc:`ValueError <stellar_sdk.exceptions.ValueError>`
        """
        return _encode_check(_VersionByte.CONTRACT, data)

    @staticmethod
    def decode_contract(data: str) -> bytes:
        """Decodes encoded contract strkey to raw data.

        :param data: encoded contract strkey
        :return: raw bytes
        :raises:
            :exc:`ValueError <stellar_sdk.exceptions.ValueError>`
        """
        try:
            return _decode_check(_VersionByte.CONTRACT, data)
        except Exception as e:
            raise ValueError(f"Invalid Pre Auth Tx Key: {data}") from e

    @staticmethod
    def is_valid_contract(contract: str) -> bool:
        """Returns ``True`` if the given `contract` is a valid encoded contract strkey.

        :param pre_auth_tx: encoded contract strkey
        :return: ``True`` if the given key is valid
        """
        return _is_valid(_VersionByte.CONTRACT, contract)


def _decode_check(version_byte: _VersionByte, encoded: str) -> bytes:
    encoded_data = encoded.encode("ascii")
    encoded_data = encoded_data + b"=" * ((8 - len(encoded_data) % 8) % 8)

    try:
        decoded_data = base64.b32decode(encoded_data)
    except binascii.Error:
        raise ValueError("Incorrect padding.")

    if encoded_data != base64.b32encode(decoded_data):  # Is that even possible?
        raise ValueError("Invalid encoded bytes.")

    version_byte_in_data = decoded_data[0:1]
    payload = decoded_data[0:-2]
    data = decoded_data[1:-2]
    checksum = decoded_data[-2:]

    if version_byte.value != version_byte_in_data:
        raise ValueError(
            f"Invalid version byte. Expected {version_byte.value!r}, got {version_byte_in_data!r}"
        )

    if version_byte == _VersionByte.ED25519_SIGNED_PAYLOAD:
        if len(data) < 32 + 4 + 4 or len(data) > 32 + 4 + 64:
            raise ValueError(f"Invalid length: {encoded}")
    elif version_byte == _VersionByte.MUXED_ACCOUNT:
        if len(data) != 32 + 8:
            raise ValueError(f"Invalid length: {encoded}")
    else:
        if len(data) != 32:
            raise ValueError(f"Invalid length: {encoded}")

    expected_checksum = _calculate_checksum(payload)
    if expected_checksum != checksum:
        raise ValueError("Invalid checksum")
    if version_byte == _VersionByte.ED25519_SIGNED_PAYLOAD:
        payload_length_prefix = int.from_bytes(data[32:36], byteorder="big")
        if len(data[36:]) % 4 != 0:
            raise ValueError(f"Invalid Ed25519 Signed Payload Key: {encoded}")
        if payload_length_prefix + ((4 - payload_length_prefix % 4) % 4) != len(
            data[36:]
        ):
            raise ValueError(f"Invalid Ed25519 Signed Payload Key: {encoded}")
    return data


def _encode_check(version_byte: _VersionByte, data: bytes) -> str:
    payload = version_byte.value + data
    crc = _calculate_checksum(payload)
    return base64.b32encode(payload + crc).decode("utf-8").rstrip("=")


def _is_valid(version_byte: _VersionByte, encoded: str) -> bool:
    try:
        _decode_check(version_byte, encoded)
    except (ValueError, TypeError):
        return False
    return True


def _get_version_byte_for_prefix(encoded: str) -> _VersionByte:
    prefix = encoded[0]
    _version_byte = ((ord(prefix) - ord("A")) << 3).to_bytes(1, byteorder="big")
    return _VersionByte(_version_byte)


def _calculate_checksum(payload: bytes) -> bytes:
    # memo note: https://gist.github.com/manran/a8357808ef71415d266dc64f0079f298
    # This code calculates CRC16-XModem checksum of payload
    checksum = binascii.crc_hqx(payload, 0)
    # Ensure that the checksum is in LSB order.
    return struct.pack("<H", checksum)



import binascii
from enum import IntEnum
from typing import Union

from . import xdr as stellar_xdr
from .strkey import StrKey
from .xdr import Hash

__all__ = ["Address"]


class AddressType(IntEnum):
    """Represents an Address type."""

    ACCOUNT = 0
    """An account address, address looks like ``GBJCHUKZMTFSLOMNC7P4TS4VJJBTCYL3XKSOLXAUJSD56C4LHND5TWUC``."""

    CONTRACT = 1
    """An contract address, address looks like ``CCJZ5DGASBWQXR5MPFCJXMBI333XE5U3FSJTNQU7RIKE3P5GN2K2WYD5``."""


class Address:
    """Represents a single address in the Stellar network.
    An address can represent an account or a contract.

    :param address: ID of the account or contract. (ex. ``GBJCHUKZMTFSLOMNC7P4TS4VJJBTCYL3XKSOLXAUJSD56C4LHND5TWUC``
        or ``CA7QYNF7SOWQ3GLR2BGMZEHXAVIRZA4KVWLTJJFC7MGXUA74P7UJUWDA``)
    """

    def __init__(self, address: str):
        if StrKey.is_valid_ed25519_public_key(address):
            self.type = AddressType.ACCOUNT
            self.key = StrKey.decode_ed25519_public_key(address)
        elif StrKey.is_valid_contract(address):
            self.type = AddressType.CONTRACT
            self.key = StrKey.decode_contract(address)
        else:
            raise ValueError("Unsupported address type.")

    @property
    def address(self) -> str:
        """Returns the encoded address.

        :return: The encoded address.
        """
        if self.type == AddressType.ACCOUNT:
            return StrKey.encode_ed25519_public_key(self.key)
        elif self.type == AddressType.CONTRACT:
            return StrKey.encode_contract(self.key)
        else:
            raise ValueError("Unsupported address type.")

    @staticmethod
    def from_raw_account(account: Union[bytes, str]) -> "Address":
        """Creates a new account Address object from raw bytes.

        :param account: The raw bytes of the account.
        :return: A new Address object.
        """
        if isinstance(account, str):
            account = binascii.unhexlify(account)
        return Address(StrKey.encode_ed25519_public_key(account))

    @staticmethod
    def from_raw_contract(contract: Union[bytes, str]) -> "Address":
        """Creates a new contract Address object from a buffer of raw bytes.

        :param contract: The raw bytes of the contract.
        :return: A new Address object.
        """
        if isinstance(contract, str):
            contract = binascii.unhexlify(contract)
        return Address(StrKey.encode_contract(contract))

    def to_xdr_sc_address(self) -> stellar_xdr.SCAddress:
        """Converts the Address object to a :class:`stellar_sdk.xdr.SCAddress` XDR object.

        :return: A :class:`stellar_sdk.xdr.SCAddress` XDR object.
        """
        if self.type == AddressType.ACCOUNT:
            account = stellar_xdr.AccountID(
                stellar_xdr.PublicKey(
                    stellar_xdr.PublicKeyType.PUBLIC_KEY_TYPE_ED25519,
                    stellar_xdr.Uint256(self.key),
                )
            )
            return stellar_xdr.SCAddress(
                stellar_xdr.SCAddressType.SC_ADDRESS_TYPE_ACCOUNT, account_id=account
            )
        elif self.type == AddressType.CONTRACT:
            contract = Hash(self.key)
            return stellar_xdr.SCAddress(
                stellar_xdr.SCAddressType.SC_ADDRESS_TYPE_CONTRACT, contract_id=contract
            )
        else:
            raise ValueError("Unsupported address type.")

    @classmethod
    def from_xdr_sc_address(cls, sc_address: stellar_xdr.SCAddress) -> "Address":
        """Creates a new Address object from a :class:`stellar_sdk.xdr.SCAddress` XDR object.

        :param sc_address: The :class:`stellar_sdk.xdr.SCAddress` XDR object.
        :return: A new Address object.
        """
        if sc_address.type == stellar_xdr.SCAddressType.SC_ADDRESS_TYPE_ACCOUNT:
            assert sc_address.account_id is not None
            assert sc_address.account_id.account_id.ed25519 is not None
            return cls.from_raw_account(
                sc_address.account_id.account_id.ed25519.uint256
            )
        elif sc_address.type == stellar_xdr.SCAddressType.SC_ADDRESS_TYPE_CONTRACT:
            assert sc_address.contract_id is not None
            return cls.from_raw_contract(sc_address.contract_id.hash)
        else:
            raise ValueError("Unsupported address type.")

    def to_xdr_sc_val(self) -> stellar_xdr.SCVal:
        return stellar_xdr.SCVal(
            stellar_xdr.SCValType.SCV_ADDRESS, address=self.to_xdr_sc_address()
        )

    @classmethod
    def from_xdr_sc_val(cls, sc_val: stellar_xdr.SCVal) -> "Address":
        if sc_val.type != stellar_xdr.SCValType.SCV_ADDRESS:
            raise ValueError("Unsupported SCVal type.")
        assert sc_val.address is not None
        return cls.from_xdr_sc_address(sc_val.address)

    def __hash__(self):
        return hash((self.key, self.type))

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.key == other.key and self.type == other.type

    def __str__(self):
        return f"<Address [type={self.type.name}, address={self.address}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64
from typing import List

from xdrlib3 import Packer, Unpacker

from .scp_envelope import SCPEnvelope
from .scp_quorum_set import SCPQuorumSet

__all__ = ["PersistedSCPStateV1"]


class PersistedSCPStateV1:
    """
    XDR Source Code::

                                                                struct PersistedSCPStateV1
                                                                {
                                                                        // Tx sets are saved separately
                                                                        SCPEnvelope scpEnvelopes<>;
                                                                        SCPQuorumSet quorumSets<>;
                                                                };
    """

    def __init__(
        self,
        scp_envelopes: List[SCPEnvelope],
        quorum_sets: List[SCPQuorumSet],
    ) -> None:
        _expect_max_length = 4294967295
        if scp_envelopes and len(scp_envelopes) > _expect_max_length:
            raise ValueError(
                f"The maximum length of `scp_envelopes` should be {_expect_max_length}, but got {len(scp_envelopes)}."
            )
        _expect_max_length = 4294967295
        if quorum_sets and len(quorum_sets) > _expect_max_length:
            raise ValueError(
                f"The maximum length of `quorum_sets` should be {_expect_max_length}, but got {len(quorum_sets)}."
            )
        self.scp_envelopes = scp_envelopes
        self.quorum_sets = quorum_sets

    def pack(self, packer: Packer) -> None:
        packer.pack_uint(len(self.scp_envelopes))
        for scp_envelopes_item in self.scp_envelopes:
            scp_envelopes_item.pack(packer)
        packer.pack_uint(len(self.quorum_sets))
        for quorum_sets_item in self.quorum_sets:
            quorum_sets_item.pack(packer)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> PersistedSCPStateV1:
        length = unpacker.unpack_uint()
        scp_envelopes = []
        for _ in range(length):
            scp_envelopes.append(SCPEnvelope.unpack(unpacker))
        length = unpacker.unpack_uint()
        quorum_sets = []
        for _ in range(length):
            quorum_sets.append(SCPQuorumSet.unpack(unpacker))
        return cls(
            scp_envelopes=scp_envelopes,
            quorum_sets=quorum_sets,
        )

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> PersistedSCPStateV1:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> PersistedSCPStateV1:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.scp_envelopes,
                self.quorum_sets,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.scp_envelopes == other.scp_envelopes
            and self.quorum_sets == other.quorum_sets
        )

    def __str__(self):
        out = [
            f"scp_envelopes={self.scp_envelopes}",
            f"quorum_sets={self.quorum_sets}",
        ]
        return f"<PersistedSCPStateV1 [{', '.join(out)}]>"



from decimal import Decimal
from typing import Any, Dict, List, Union

from .account import Account
from .asset import Asset
from .base_server import BaseServer
from .call_builder.call_builder_sync import *
from .client.base_sync_client import BaseSyncClient
from .client.requests_client import RequestsClient
from .client.response import Response
from .exceptions import NotFoundError, raise_request_exception
from .fee_bump_transaction import FeeBumpTransaction
from .fee_bump_transaction_envelope import FeeBumpTransactionEnvelope
from .keypair import Keypair
from .memo import NoneMemo
from .muxed_account import MuxedAccount
from .transaction import Transaction
from .transaction_envelope import TransactionEnvelope
from .utils import MUXED_ACCOUNT_STARTING_LETTER, urljoin_with_query

__all__ = ["Server"]


class Server(BaseServer):
    """Server handles the network connection to a `Horizon <https://developers.stellar.org/api/introduction/>`_
    instance and exposes an interface for requests to that instance.

    An example::

        from stellar_sdk import Server

        server = Server("https://horizon-testnet.stellar.org")
        resp = server.transactions().limit(10).order(desc=True).call()
        print(resp)

    :param horizon_url: Horizon Server URL
        (ex. ``"https://horizon-testnet.stellar.org"`` for test network,
        ``"https://horizon.stellar.org"`` for public network)
    :param client: Http client used to send the request
    """

    def __init__(
        self,
        horizon_url: str = "https://horizon-testnet.stellar.org/",
        client: BaseSyncClient = None,
    ) -> None:
        self.horizon_url: str = horizon_url

        if not client:
            client = RequestsClient()
        self._client: BaseSyncClient = client

    def submit_transaction(
        self,
        transaction_envelope: Union[
            TransactionEnvelope, FeeBumpTransactionEnvelope, str
        ],
        skip_memo_required_check: bool = False,
    ) -> Dict[str, Any]:
        """Submits a transaction to the network.

        :param transaction_envelope: :class:`stellar_sdk.transaction_envelope.TransactionEnvelope` object
            or base64 encoded xdr
        :param skip_memo_required_check: Allow skipping memo
        :return: the response from horizon
        :raises:
            :exc:`ConnectionError <stellar_sdk.exceptions.ConnectionError>`
            :exc:`NotFoundError <stellar_sdk.exceptions.NotFoundError>`
            :exc:`BadRequestError <stellar_sdk.exceptions.BadRequestError>`
            :exc:`BadResponseError <stellar_sdk.exceptions.BadResponseError>`
            :exc:`UnknownRequestError <stellar_sdk.exceptions.UnknownRequestError>`
            :exc:`AccountRequiresMemoError <stellar_sdk.sep.exceptions.AccountRequiresMemoError>`
        """
        url = urljoin_with_query(self.horizon_url, "transactions")
        xdr, tx = self._get_xdr_and_transaction_from_transaction_envelope(
            transaction_envelope
        )
        if not skip_memo_required_check:
            self.__check_memo_required_sync(tx)
        data = {"tx": xdr}
        resp = self._client.post(url=url, data=data)
        assert isinstance(resp, Response)
        raise_request_exception(resp)
        return resp.json()

    def load_account(self, account_id: Union[MuxedAccount, Keypair, str]) -> Account:
        """Fetches an account's most current base state (like sequence) in the ledger and then creates
        and returns an :class:`stellar_sdk.account.Account` object.

        If you want to get complete account information, please
        use :func:`stellar_sdk.server.Server.accounts`.

        :param account_id: The account to load.
        :return: an :class:`stellar_sdk.account.Account` object.
        :raises:
            :exc:`ConnectionError <stellar_sdk.exceptions.ConnectionError>`
            :exc:`NotFoundError <stellar_sdk.exceptions.NotFoundError>`
            :exc:`BadRequestError <stellar_sdk.exceptions.BadRequestError>`
            :exc:`BadResponseError <stellar_sdk.exceptions.BadResponseError>`
            :exc:`UnknownRequestError <stellar_sdk.exceptions.UnknownRequestError>`
        """
        if isinstance(account_id, str):
            account_id = MuxedAccount.from_account(account_id)
        elif isinstance(account_id, Keypair):
            account_id = MuxedAccount.from_account(account_id.public_key)
        else:
            account_id = account_id
        resp = self.accounts().account_id(account_id=account_id.account_id).call()
        sequence = int(resp["sequence"])
        account = Account(account=account_id, sequence=sequence, raw_data=resp)
        return account

    def __check_memo_required_sync(
        self, transaction: Union[Transaction, FeeBumpTransaction]
    ) -> None:
        if isinstance(transaction, FeeBumpTransaction):
            transaction = transaction.inner_transaction_envelope.transaction
        if not (transaction.memo is None or isinstance(transaction.memo, NoneMemo)):
            return
        for index, destination in self._get_check_memo_required_destinations(
            transaction
        ):
            if destination.startswith(MUXED_ACCOUNT_STARTING_LETTER):
                continue
            try:
                account_resp = self.accounts().account_id(destination).call()
            except NotFoundError:
                continue
            assert isinstance(account_resp, dict)
            self._check_destination_memo(account_resp, index, destination)

    def fetch_base_fee(self) -> int:
        """Fetch the base fee. Since this hits the server, if the server call fails,
        you might get an error. You should be prepared to use a default value if that happens.

        :return: the base fee
        :raises:
            :exc:`ConnectionError <stellar_sdk.exceptions.ConnectionError>`
            :exc:`NotFoundError <stellar_sdk.exceptions.NotFoundError>`
            :exc:`BadRequestError <stellar_sdk.exceptions.BadRequestError>`
            :exc:`BadResponseError <stellar_sdk.exceptions.BadResponseError>`
            :exc:`UnknownRequestError <stellar_sdk.exceptions.UnknownRequestError>`
        """
        latest_ledger = self.ledgers().order(desc=True).limit(1).call()
        assert isinstance(latest_ledger, dict)
        base_fee = self._handle_base_fee(latest_ledger)
        return base_fee

    def root(self) -> RootCallBuilder:
        """
        :return: New :class:`stellar_sdk.call_builder.call_builder_sync.RootCallBuilder` object configured
            by a current Horizon server configuration.
        """
        return RootCallBuilder(horizon_url=self.horizon_url, client=self._client)

    def accounts(self) -> AccountsCallBuilder:
        """
        :return: New :class:`stellar_sdk.call_builder.call_builder_sync.AccountsCallBuilder` object configured
            by a current Horizon server configuration.
        """
        return AccountsCallBuilder(horizon_url=self.horizon_url, client=self._client)

    def assets(self) -> AssetsCallBuilder:
        """
        :return: New :class:`stellar_sdk.call_builder.call_builder_sync.AssetsCallBuilder` object configured by
            a current Horizon server configuration.
        """
        return AssetsCallBuilder(horizon_url=self.horizon_url, client=self._client)

    def claimable_balances(self) -> ClaimableBalancesCallBuilder:
        """
        :return: New :class:`stellar_sdk.call_builder.call_builder_sync.ClaimableBalancesCallBuilder` object configured by
            a current Horizon server configuration.
        """
        return ClaimableBalancesCallBuilder(
            horizon_url=self.horizon_url, client=self._client
        )

    def data(self, account_id: str, data_name: str) -> DataCallBuilder:
        """
        :return: New :class:`stellar_sdk.call_builder.call_builder_sync.DataCallBuilder` object configured by
            a current Horizon server configuration.
        """
        return DataCallBuilder(
            horizon_url=self.horizon_url,
            client=self._client,
            account_id=account_id,
            data_name=data_name,
        )

    def effects(self) -> EffectsCallBuilder:
        """
        :return: New :class:`stellar_sdk.call_builder.call_builder_sync.EffectsCallBuilder` object configured by
            a current Horizon server configuration.
        """
        return EffectsCallBuilder(horizon_url=self.horizon_url, client=self._client)

    def fee_stats(self) -> FeeStatsCallBuilder:
        """
        :return: New :class:`stellar_sdk.call_builder.call_builder_sync.FeeStatsCallBuilder` object configured by
            a current Horizon server configuration.
        """
        return FeeStatsCallBuilder(horizon_url=self.horizon_url, client=self._client)

    def ledgers(self) -> LedgersCallBuilder:
        """
        :return: New :class:`stellar_sdk.call_builder.call_builder_sync.LedgersCallBuilder` object configured by
            a current Horizon server configuration.
        """
        return LedgersCallBuilder(horizon_url=self.horizon_url, client=self._client)

    def liquidity_pools(self) -> LiquidityPoolsBuilder:
        """
        :return: New :class:`stellar_sdk.call_builder.call_builder_sync.LiquidityPoolsBuilder` object configured by
            a current Horizon server configuration.
        """
        return LiquidityPoolsBuilder(horizon_url=self.horizon_url, client=self._client)

    def offers(self) -> OffersCallBuilder:
        """
        :return: New :class:`stellar_sdk.call_builder.call_builder_sync.OffersCallBuilder` object configured by
            a current Horizon server configuration.
        """
        return OffersCallBuilder(horizon_url=self.horizon_url, client=self._client)

    def operations(self) -> OperationsCallBuilder:
        """
        :return: New :class:`stellar_sdk.call_builder.call_builder_sync.OperationsCallBuilder` object configured by
            a current Horizon server configuration.
        """
        return OperationsCallBuilder(horizon_url=self.horizon_url, client=self._client)

    def orderbook(self, selling: Asset, buying: Asset) -> OrderbookCallBuilder:
        """
        :param selling: Asset being sold
        :param buying: Asset being bought
        :return: New :class:`stellar_sdk.call_builder.call_builder_sync.OrderbookCallBuilder` object configured by
            a current Horizon server configuration.
        """
        return OrderbookCallBuilder(
            horizon_url=self.horizon_url,
            client=self._client,
            buying=buying,
            selling=selling,
        )

    def strict_receive_paths(
        self,
        source: Union[str, List[Asset]],
        destination_asset: Asset,
        destination_amount: Union[str, Decimal],
    ) -> StrictReceivePathsCallBuilder:
        """
        :param source: The sender's account ID or a list of Assets. Any returned path must use a source that the sender can hold.
        :param destination_asset: The destination asset.
        :param destination_amount: The amount, denominated in the destination asset, that any returned path should be able to satisfy.
        :return: New :class:`stellar_sdk.call_builder.call_builder_sync.StrictReceivePathsCallBuilder` object configured by
            a current Horizon server configuration.
        """
        return StrictReceivePathsCallBuilder(
            horizon_url=self.horizon_url,
            client=self._client,
            source=source,
            destination_asset=destination_asset,
            destination_amount=destination_amount,
        )

    def strict_send_paths(
        self,
        source_asset: Asset,
        source_amount: Union[str, Decimal],
        destination: Union[str, List[Asset]],
    ) -> StrictSendPathsCallBuilder:
        """
        :param source_asset: The asset to be sent.
        :param source_amount: The amount, denominated in the source asset, that any returned path should be able to satisfy.
        :param destination: The destination account or the destination assets.
        :return: New :class:`stellar_sdk.call_builder.call_builder_sync.StrictReceivePathsCallBuilder` object configured by
            a current Horizon server configuration.
        """
        return StrictSendPathsCallBuilder(
            horizon_url=self.horizon_url,
            client=self._client,
            source_asset=source_asset,
            source_amount=source_amount,
            destination=destination,
        )

    def payments(self) -> PaymentsCallBuilder:
        """
        :return: New :class:`stellar_sdk.call_builder.call_builder_sync.PaymentsCallBuilder` object configured by
            a current Horizon server configuration.
        """
        return PaymentsCallBuilder(horizon_url=self.horizon_url, client=self._client)

    def trade_aggregations(
        self,
        base: Asset,
        counter: Asset,
        resolution: int,
        start_time: int = None,
        end_time: int = None,
        offset: int = None,
    ) -> TradeAggregationsCallBuilder:
        """
        :param base: base asset
        :param counter: counter asset
        :param resolution: segment duration as millis since epoch. *Supported values
            are 1 minute (60000), 5 minutes (300000), 15 minutes (900000),
            1 hour (3600000), 1 day (86400000) and 1 week (604800000).*
        :param start_time: lower time boundary represented as millis since epoch
        :param end_time: upper time boundary represented as millis since epoch
        :param offset: segments can be offset using this parameter.
            Expressed in milliseconds. *Can only be used if the resolution is greater than 1 hour.
            Value must be in whole hours, less than the provided resolution, and less than 24 hours.*
        :return: New :class:`stellar_sdk.call_builder.call_builder_sync.TradeAggregationsCallBuilder` object configured by
            a current Horizon server configuration.
        """
        return TradeAggregationsCallBuilder(
            horizon_url=self.horizon_url,
            client=self._client,
            base=base,
            counter=counter,
            start_time=start_time,
            end_time=end_time,
            resolution=resolution,
            offset=offset,
        )

    def trades(self) -> TradesCallBuilder:
        """
        :return: New :class:`stellar_sdk.call_builder.call_builder_sync.TradesCallBuilder` object configured by
            a current Horizon server configuration.
        """
        return TradesCallBuilder(horizon_url=self.horizon_url, client=self._client)

    def transactions(self) -> TransactionsCallBuilder:
        """
        :return: New :class:`stellar_sdk.call_builder.call_builder_sync.TransactionsCallBuilder` object configured by
            a current Horizon server configuration.
        """
        return TransactionsCallBuilder(
            horizon_url=self.horizon_url, client=self._client
        )

    def close(self) -> None:
        """Close underlying connector, and release all acquired resources."""
        self._client.close()

    def __enter__(self) -> "Server":
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def __str__(self):
        return f"<Server [horizon_url={self.horizon_url}, client={self._client}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64

from xdrlib3 import Packer, Unpacker

from .precondition_type import PreconditionType
from .preconditions_v2 import PreconditionsV2
from .time_bounds import TimeBounds

__all__ = ["Preconditions"]


class Preconditions:
    """
    XDR Source Code::

        union Preconditions switch (PreconditionType type)
        {
        case PRECOND_NONE:
            void;
        case PRECOND_TIME:
            TimeBounds timeBounds;
        case PRECOND_V2:
            PreconditionsV2 v2;
        };
    """

    def __init__(
        self,
        type: PreconditionType,
        time_bounds: TimeBounds = None,
        v2: PreconditionsV2 = None,
    ) -> None:
        self.type = type
        self.time_bounds = time_bounds
        self.v2 = v2

    def pack(self, packer: Packer) -> None:
        self.type.pack(packer)
        if self.type == PreconditionType.PRECOND_NONE:
            return
        if self.type == PreconditionType.PRECOND_TIME:
            if self.time_bounds is None:
                raise ValueError("time_bounds should not be None.")
            self.time_bounds.pack(packer)
            return
        if self.type == PreconditionType.PRECOND_V2:
            if self.v2 is None:
                raise ValueError("v2 should not be None.")
            self.v2.pack(packer)
            return

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> Preconditions:
        type = PreconditionType.unpack(unpacker)
        if type == PreconditionType.PRECOND_NONE:
            return cls(type=type)
        if type == PreconditionType.PRECOND_TIME:
            time_bounds = TimeBounds.unpack(unpacker)
            return cls(type=type, time_bounds=time_bounds)
        if type == PreconditionType.PRECOND_V2:
            v2 = PreconditionsV2.unpack(unpacker)
            return cls(type=type, v2=v2)
        return cls(type=type)

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> Preconditions:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> Preconditions:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.type,
                self.time_bounds,
                self.v2,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.type == other.type
            and self.time_bounds == other.time_bounds
            and self.v2 == other.v2
        )

    def __str__(self):
        out = []
        out.append(f"type={self.type}")
        out.append(
            f"time_bounds={self.time_bounds}"
        ) if self.time_bounds is not None else None
        out.append(f"v2={self.v2}") if self.v2 is not None else None
        return f"<Preconditions [{', '.join(out)}]>"



from typing import Union

from .fee_bump_transaction_envelope import FeeBumpTransactionEnvelope
from .transaction_envelope import TransactionEnvelope

__all__ = ["parse_transaction_envelope_from_xdr"]


def parse_transaction_envelope_from_xdr(
    xdr: str, network_passphrase: str
) -> Union[TransactionEnvelope, FeeBumpTransactionEnvelope]:
    """When you are not sure whether your XDR belongs to
    :py:class:`TransactionEnvelope <stellar_sdk.transaction_envelope.TransactionEnvelope>`
    or :py:class:`FeeBumpTransactionEnvelope <stellar_sdk.fee_bump_transaction_envelope.FeeBumpTransactionEnvelope>`,
    you can use this helper function.

    An example::

        from stellar_sdk import Network
        from stellar_sdk.helpers import parse_transaction_envelope_from_xdr

        xdr = "AAAAAgAAAADHJNEDn33/C1uDkDfzDfKVq/4XE9IxDfGiLCfoV7riZQAAA+gCI4TVABpRPgAAAAAAAAAAAAAAAQAAAAAAAAADAAAAAUxpcmEAAAAAabIaDgm0ypyJpsVfEjZw2mO3Enq4Q4t5URKfWtqukSUAAAABVVNEAAAAAADophqGHmCvYPgHc+BjRuXHLL5Z3K3aN2CNWO9CUR2f3AAAAAAAAAAAE8G9mAADcH8AAAAAMYdBWgAAAAAAAAABV7riZQAAAEARGCGwYk/kEB2Z4UL20y536evnwmmSc4c2FnxlvUcPZl5jgWHcNwY8LTpFhdrUN9TZWciCRp/JCZYa0SJh8cYB"
        te = parse_transaction_envelope_from_xdr(xdr, Network.PUBLIC_NETWORK_PASSPHRASE)
        print(te)

    :param xdr: Transaction envelope XDR
    :param network_passphrase: The network to connect to for verifying and retrieving
        additional attributes from. (ex. ``"Public Global Stellar Network ; September 2015"``)
    :raises: :exc:`ValueError <stellar_sdk.exceptions.ValueError>` - XDR is neither :py:class:`TransactionEnvelope <stellar_sdk.transaction_envelope.TransactionEnvelope>`
        nor :py:class:`FeeBumpTransactionEnvelope <stellar_sdk.fee_bump_transaction_envelope.FeeBumpTransactionEnvelope>`
    """
    if FeeBumpTransactionEnvelope.is_fee_bump_transaction_envelope(xdr):
        return FeeBumpTransactionEnvelope.from_xdr(xdr, network_passphrase)
    return TransactionEnvelope.from_xdr(xdr, network_passphrase)



import warnings
from enum import IntFlag
from typing import Optional, Union

from .. import xdr as stellar_xdr
from ..asset import Asset
from ..keypair import Keypair
from ..muxed_account import MuxedAccount
from ..strkey import StrKey
from ..utils import raise_if_not_valid_ed25519_public_key
from .operation import Operation

__all__ = ["TrustLineEntryFlag", "AllowTrust"]


class TrustLineEntryFlag(IntFlag):
    """Indicates which flags to set. For details about the flags,
    please refer to the `CAP-0018 <https://github.com/stellar/stellar-protocol/blob/master/core/cap-0018.md>`_.

    - **UNAUTHORIZED_FLAG**: The account can hold a balance but cannot receive payments, send payments, maintain offers or manage offers

    - **AUTHORIZED_FLAG**: The account can hold a balance, receive payments, send payments, maintain offers or manage offers

    - **AUTHORIZED_TO_MAINTAIN_LIABILITIES_FLAG**: The account can hold a balance and maintain offers but cannot receive payments, send payments or manage offers
    """

    UNAUTHORIZED_FLAG = 0
    AUTHORIZED_FLAG = 1
    AUTHORIZED_TO_MAINTAIN_LIABILITIES_FLAG = 2


class AllowTrust(Operation):
    """The :class:`AllowTrust` object, which represents a AllowTrust operation
    on Stellar's network.

    Updates the authorized flag of an existing trustline. This can only be
    called by the issuer of a trustline's `asset
    <https://developers.stellar.org/docs/glossary/assets/>`_.

    The issuer can only clear the authorized flag if the issuer has the
    ``AUTH_REVOCABLE_FLAG`` set. Otherwise, the issuer can only set the authorized
    flag.

    Threshold: Low

    See `Allow Trust <https://developers.stellar.org/docs/start/list-of-operations/#allow-trust>`_ for more information.

    :param trustor: The trusting account (the one being authorized).
    :param asset_code: The asset code being authorized.
    :param authorize: `True` to authorize the line, `False` to deauthorize，if you need further control,
        you can also use :class:`stellar_sdk.operation.allow_trust.TrustLineEntryFlag`.
    :param source: The source account for the operation. Defaults to the transaction's source account.

    """

    _XDR_OPERATION_TYPE: stellar_xdr.OperationType = (
        stellar_xdr.OperationType.ALLOW_TRUST
    )

    def __init__(
        self,
        trustor: str,
        asset_code: str,
        authorize: Union[TrustLineEntryFlag, bool],
        source: Optional[Union[MuxedAccount, str]] = None,
    ) -> None:
        # We keep this class to ensure that the SDK can correctly parse historical transactions.
        warnings.warn(
            "Use `stellar_sdk.operation.set_trust_line_flags.SetTrustLineFlags` instead.",
            DeprecationWarning,
        )
        super().__init__(source)

        self.trustor: str = trustor
        self.asset_code: str = asset_code

        if isinstance(authorize, bool):
            if authorize is True:
                self.authorize: TrustLineEntryFlag = TrustLineEntryFlag.AUTHORIZED_FLAG
            else:
                self.authorize = TrustLineEntryFlag.UNAUTHORIZED_FLAG
        else:
            self.authorize = authorize
        raise_if_not_valid_ed25519_public_key(self.trustor, "trustor")
        Asset.check_if_asset_code_is_valid(self.asset_code)

    def _to_operation_body(self) -> stellar_xdr.OperationBody:
        trustor = Keypair.from_public_key(self.trustor).xdr_account_id()
        length = len(self.asset_code)
        pad_length = 4 - length if length <= 4 else 12 - length
        asset_code = bytearray(self.asset_code, "ascii") + b"\x00" * pad_length
        authorize = stellar_xdr.Uint32(self.authorize.value)
        if len(asset_code) == 4:
            asset_type = stellar_xdr.AssetType.ASSET_TYPE_CREDIT_ALPHANUM4
            asset_code4 = stellar_xdr.AssetCode4(asset_code)
            asset = stellar_xdr.AssetCode(type=asset_type, asset_code4=asset_code4)
        else:
            asset_type = stellar_xdr.AssetType.ASSET_TYPE_CREDIT_ALPHANUM12
            asset_code12 = stellar_xdr.AssetCode12(asset_code)
            asset = stellar_xdr.AssetCode(type=asset_type, asset_code12=asset_code12)
        allow_trust_op = stellar_xdr.AllowTrustOp(trustor, asset, authorize)
        body = stellar_xdr.OperationBody(
            type=self._XDR_OPERATION_TYPE, allow_trust_op=allow_trust_op
        )
        return body

    @classmethod
    def from_xdr_object(cls, xdr_object: stellar_xdr.Operation) -> "AllowTrust":
        """Creates a :class:`AllowTrust` object from an XDR Operation
        object.

        """
        source = Operation.get_source_from_xdr_obj(xdr_object)
        assert xdr_object.body.allow_trust_op is not None
        assert xdr_object.body.allow_trust_op.trustor.account_id.ed25519 is not None
        trustor = StrKey.encode_ed25519_public_key(
            xdr_object.body.allow_trust_op.trustor.account_id.ed25519.uint256
        )
        authorize = xdr_object.body.allow_trust_op.authorize.uint32
        authorize = TrustLineEntryFlag(authorize)
        asset_type = xdr_object.body.allow_trust_op.asset.type
        if asset_type == stellar_xdr.AssetType.ASSET_TYPE_CREDIT_ALPHANUM4:
            assert xdr_object.body.allow_trust_op.asset.asset_code4 is not None
            asset_code = (
                xdr_object.body.allow_trust_op.asset.asset_code4.asset_code4.decode()
            )
        elif asset_type == stellar_xdr.AssetType.ASSET_TYPE_CREDIT_ALPHANUM12:
            assert xdr_object.body.allow_trust_op.asset.asset_code12 is not None
            asset_code = (
                xdr_object.body.allow_trust_op.asset.asset_code12.asset_code12.decode()
            )
        else:
            raise NotImplementedError(
                "Operation of asset_type={} is not implemented" ".".format(asset_type)
            )

        asset_code = asset_code.rstrip("\x00")
        op = cls(
            source=source, trustor=trustor, authorize=authorize, asset_code=asset_code
        )
        return op

    def __str__(self):
        return (
            f"<AllowTrust [trustor={self.trustor}, "
            f"asset_code={self.asset_code}, "
            f"authorize={self.authorize}, "
            f"source={self.source}]>"
        )



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64

from xdrlib3 import Packer, Unpacker

from .scp_nomination import SCPNomination
from .scp_statement_confirm import SCPStatementConfirm
from .scp_statement_externalize import SCPStatementExternalize
from .scp_statement_prepare import SCPStatementPrepare
from .scp_statement_type import SCPStatementType

__all__ = ["SCPStatementPledges"]


class SCPStatementPledges:
    """
    XDR Source Code::

        union switch (SCPStatementType type)
            {
            case SCP_ST_PREPARE:
                struct
                {
                    Hash quorumSetHash;       // D
                    SCPBallot ballot;         // b
                    SCPBallot* prepared;      // p
                    SCPBallot* preparedPrime; // p'
                    uint32 nC;                // c.n
                    uint32 nH;                // h.n
                } prepare;
            case SCP_ST_CONFIRM:
                struct
                {
                    SCPBallot ballot;   // b
                    uint32 nPrepared;   // p.n
                    uint32 nCommit;     // c.n
                    uint32 nH;          // h.n
                    Hash quorumSetHash; // D
                } confirm;
            case SCP_ST_EXTERNALIZE:
                struct
                {
                    SCPBallot commit;         // c
                    uint32 nH;                // h.n
                    Hash commitQuorumSetHash; // D used before EXTERNALIZE
                } externalize;
            case SCP_ST_NOMINATE:
                SCPNomination nominate;
            }
    """

    def __init__(
        self,
        type: SCPStatementType,
        prepare: SCPStatementPrepare = None,
        confirm: SCPStatementConfirm = None,
        externalize: SCPStatementExternalize = None,
        nominate: SCPNomination = None,
    ) -> None:
        self.type = type
        self.prepare = prepare
        self.confirm = confirm
        self.externalize = externalize
        self.nominate = nominate

    def pack(self, packer: Packer) -> None:
        self.type.pack(packer)
        if self.type == SCPStatementType.SCP_ST_PREPARE:
            if self.prepare is None:
                raise ValueError("prepare should not be None.")
            self.prepare.pack(packer)
            return
        if self.type == SCPStatementType.SCP_ST_CONFIRM:
            if self.confirm is None:
                raise ValueError("confirm should not be None.")
            self.confirm.pack(packer)
            return
        if self.type == SCPStatementType.SCP_ST_EXTERNALIZE:
            if self.externalize is None:
                raise ValueError("externalize should not be None.")
            self.externalize.pack(packer)
            return
        if self.type == SCPStatementType.SCP_ST_NOMINATE:
            if self.nominate is None:
                raise ValueError("nominate should not be None.")
            self.nominate.pack(packer)
            return

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> SCPStatementPledges:
        type = SCPStatementType.unpack(unpacker)
        if type == SCPStatementType.SCP_ST_PREPARE:
            prepare = SCPStatementPrepare.unpack(unpacker)
            return cls(type=type, prepare=prepare)
        if type == SCPStatementType.SCP_ST_CONFIRM:
            confirm = SCPStatementConfirm.unpack(unpacker)
            return cls(type=type, confirm=confirm)
        if type == SCPStatementType.SCP_ST_EXTERNALIZE:
            externalize = SCPStatementExternalize.unpack(unpacker)
            return cls(type=type, externalize=externalize)
        if type == SCPStatementType.SCP_ST_NOMINATE:
            nominate = SCPNomination.unpack(unpacker)
            return cls(type=type, nominate=nominate)
        return cls(type=type)

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> SCPStatementPledges:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> SCPStatementPledges:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.type,
                self.prepare,
                self.confirm,
                self.externalize,
                self.nominate,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.type == other.type
            and self.prepare == other.prepare
            and self.confirm == other.confirm
            and self.externalize == other.externalize
            and self.nominate == other.nominate
        )

    def __str__(self):
        out = []
        out.append(f"type={self.type}")
        out.append(f"prepare={self.prepare}") if self.prepare is not None else None
        out.append(f"confirm={self.confirm}") if self.confirm is not None else None
        out.append(
            f"externalize={self.externalize}"
        ) if self.externalize is not None else None
        out.append(f"nominate={self.nominate}") if self.nominate is not None else None
        return f"<SCPStatementPledges [{', '.join(out)}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64
from enum import IntEnum

from xdrlib3 import Packer, Unpacker

__all__ = ["TransactionResultCode"]


class TransactionResultCode(IntEnum):
    """
    XDR Source Code::

        enum TransactionResultCode
        {
            txFEE_BUMP_INNER_SUCCESS = 1, // fee bump inner transaction succeeded
            txSUCCESS = 0,                // all operations succeeded

            txFAILED = -1, // one of the operations failed (none were applied)

            txTOO_EARLY = -2,         // ledger closeTime before minTime
            txTOO_LATE = -3,          // ledger closeTime after maxTime
            txMISSING_OPERATION = -4, // no operation was specified
            txBAD_SEQ = -5,           // sequence number does not match source account

            txBAD_AUTH = -6,             // too few valid signatures / wrong network
            txINSUFFICIENT_BALANCE = -7, // fee would bring account below reserve
            txNO_ACCOUNT = -8,           // source account not found
            txINSUFFICIENT_FEE = -9,     // fee is too small
            txBAD_AUTH_EXTRA = -10,      // unused signatures attached to transaction
            txINTERNAL_ERROR = -11,      // an unknown error occurred

            txNOT_SUPPORTED = -12,          // transaction type not supported
            txFEE_BUMP_INNER_FAILED = -13,  // fee bump inner transaction failed
            txBAD_SPONSORSHIP = -14,        // sponsorship not confirmed
            txBAD_MIN_SEQ_AGE_OR_GAP = -15, // minSeqAge or minSeqLedgerGap conditions not met
            txMALFORMED = -16,              // precondition is invalid
            txSOROBAN_INVALID = -17         // soroban-specific preconditions were not met
        };
    """

    txFEE_BUMP_INNER_SUCCESS = 1
    txSUCCESS = 0
    txFAILED = -1
    txTOO_EARLY = -2
    txTOO_LATE = -3
    txMISSING_OPERATION = -4
    txBAD_SEQ = -5
    txBAD_AUTH = -6
    txINSUFFICIENT_BALANCE = -7
    txNO_ACCOUNT = -8
    txINSUFFICIENT_FEE = -9
    txBAD_AUTH_EXTRA = -10
    txINTERNAL_ERROR = -11
    txNOT_SUPPORTED = -12
    txFEE_BUMP_INNER_FAILED = -13
    txBAD_SPONSORSHIP = -14
    txBAD_MIN_SEQ_AGE_OR_GAP = -15
    txMALFORMED = -16
    txSOROBAN_INVALID = -17

    def pack(self, packer: Packer) -> None:
        packer.pack_int(self.value)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> TransactionResultCode:
        value = unpacker.unpack_int()
        return cls(value)

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> TransactionResultCode:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> TransactionResultCode:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64

from xdrlib3 import Packer, Unpacker

from .sc_spec_type_def import SCSpecTypeDef

__all__ = ["SCSpecTypeResult"]


class SCSpecTypeResult:
    """
    XDR Source Code::

        struct SCSpecTypeResult
        {
            SCSpecTypeDef okType;
            SCSpecTypeDef errorType;
        };
    """

    def __init__(
        self,
        ok_type: SCSpecTypeDef,
        error_type: SCSpecTypeDef,
    ) -> None:
        self.ok_type = ok_type
        self.error_type = error_type

    def pack(self, packer: Packer) -> None:
        self.ok_type.pack(packer)
        self.error_type.pack(packer)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> SCSpecTypeResult:
        ok_type = SCSpecTypeDef.unpack(unpacker)
        error_type = SCSpecTypeDef.unpack(unpacker)
        return cls(
            ok_type=ok_type,
            error_type=error_type,
        )

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> SCSpecTypeResult:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> SCSpecTypeResult:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.ok_type,
                self.error_type,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.ok_type == other.ok_type and self.error_type == other.error_type

    def __str__(self):
        out = [
            f"ok_type={self.ok_type}",
            f"error_type={self.error_type}",
        ]
        return f"<SCSpecTypeResult [{', '.join(out)}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64
from enum import IntEnum

from xdrlib3 import Packer, Unpacker

__all__ = ["PathPaymentStrictReceiveResultCode"]


class PathPaymentStrictReceiveResultCode(IntEnum):
    """
    XDR Source Code::

        enum PathPaymentStrictReceiveResultCode
        {
            // codes considered as "success" for the operation
            PATH_PAYMENT_STRICT_RECEIVE_SUCCESS = 0, // success

            // codes considered as "failure" for the operation
            PATH_PAYMENT_STRICT_RECEIVE_MALFORMED = -1, // bad input
            PATH_PAYMENT_STRICT_RECEIVE_UNDERFUNDED =
                -2, // not enough funds in source account
            PATH_PAYMENT_STRICT_RECEIVE_SRC_NO_TRUST =
                -3, // no trust line on source account
            PATH_PAYMENT_STRICT_RECEIVE_SRC_NOT_AUTHORIZED =
                -4, // source not authorized to transfer
            PATH_PAYMENT_STRICT_RECEIVE_NO_DESTINATION =
                -5, // destination account does not exist
            PATH_PAYMENT_STRICT_RECEIVE_NO_TRUST =
                -6, // dest missing a trust line for asset
            PATH_PAYMENT_STRICT_RECEIVE_NOT_AUTHORIZED =
                -7, // dest not authorized to hold asset
            PATH_PAYMENT_STRICT_RECEIVE_LINE_FULL =
                -8, // dest would go above their limit
            PATH_PAYMENT_STRICT_RECEIVE_NO_ISSUER = -9, // missing issuer on one asset
            PATH_PAYMENT_STRICT_RECEIVE_TOO_FEW_OFFERS =
                -10, // not enough offers to satisfy path
            PATH_PAYMENT_STRICT_RECEIVE_OFFER_CROSS_SELF =
                -11, // would cross one of its own offers
            PATH_PAYMENT_STRICT_RECEIVE_OVER_SENDMAX = -12 // could not satisfy sendmax
        };
    """

    PATH_PAYMENT_STRICT_RECEIVE_SUCCESS = 0
    PATH_PAYMENT_STRICT_RECEIVE_MALFORMED = -1
    PATH_PAYMENT_STRICT_RECEIVE_UNDERFUNDED = -2
    PATH_PAYMENT_STRICT_RECEIVE_SRC_NO_TRUST = -3
    PATH_PAYMENT_STRICT_RECEIVE_SRC_NOT_AUTHORIZED = -4
    PATH_PAYMENT_STRICT_RECEIVE_NO_DESTINATION = -5
    PATH_PAYMENT_STRICT_RECEIVE_NO_TRUST = -6
    PATH_PAYMENT_STRICT_RECEIVE_NOT_AUTHORIZED = -7
    PATH_PAYMENT_STRICT_RECEIVE_LINE_FULL = -8
    PATH_PAYMENT_STRICT_RECEIVE_NO_ISSUER = -9
    PATH_PAYMENT_STRICT_RECEIVE_TOO_FEW_OFFERS = -10
    PATH_PAYMENT_STRICT_RECEIVE_OFFER_CROSS_SELF = -11
    PATH_PAYMENT_STRICT_RECEIVE_OVER_SENDMAX = -12

    def pack(self, packer: Packer) -> None:
        packer.pack_int(self.value)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> PathPaymentStrictReceiveResultCode:
        value = unpacker.unpack_int()
        return cls(value)

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> PathPaymentStrictReceiveResultCode:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> PathPaymentStrictReceiveResultCode:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64

from xdrlib3 import Packer, Unpacker

from .claimable_balance_entry_extension_v1_ext import (
    ClaimableBalanceEntryExtensionV1Ext,
)
from .uint32 import Uint32

__all__ = ["ClaimableBalanceEntryExtensionV1"]


class ClaimableBalanceEntryExtensionV1:
    """
    XDR Source Code::

        struct ClaimableBalanceEntryExtensionV1
        {
            union switch (int v)
            {
            case 0:
                void;
            }
            ext;

            uint32 flags; // see ClaimableBalanceFlags
        };
    """

    def __init__(
        self,
        ext: ClaimableBalanceEntryExtensionV1Ext,
        flags: Uint32,
    ) -> None:
        self.ext = ext
        self.flags = flags

    def pack(self, packer: Packer) -> None:
        self.ext.pack(packer)
        self.flags.pack(packer)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> ClaimableBalanceEntryExtensionV1:
        ext = ClaimableBalanceEntryExtensionV1Ext.unpack(unpacker)
        flags = Uint32.unpack(unpacker)
        return cls(
            ext=ext,
            flags=flags,
        )

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> ClaimableBalanceEntryExtensionV1:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> ClaimableBalanceEntryExtensionV1:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.ext,
                self.flags,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.ext == other.ext and self.flags == other.flags

    def __str__(self):
        out = [
            f"ext={self.ext}",
            f"flags={self.flags}",
        ]
        return f"<ClaimableBalanceEntryExtensionV1 [{', '.join(out)}]>"



from typing import Union

from . import xdr as stellar_xdr
from .signer_key import SignerKey

__all__ = ["Signer"]


class Signer:
    """The :class:`Signer` object, which represents an account signer on Stellar's network.

    An example:

        from stellar_sdk import Signer

        signer_ed25519 = Signer.ed25519_public_key("GCC3U63F5OJIG4VS6XCFUJGCQRRMNCVGASDGIZZEPA3AZ242K4JVPIYV", 1)
        signer_sha256_hash = Signer.sha256_hash("XCC3U63F5OJIG4VS6XCFUJGCQRRMNCVGASDGIZZEPA3AZ242K4JVPRP5", 2)
        signer_pre_auth_tx = Signer.pre_auth_tx("TCC3U63F5OJIG4VS6XCFUJGCQRRMNCVGASDGIZZEPA3AZ242K4JVOVKE", 3)
        print(f"signer_ed25519 account id: {signer_ed25519.signer_key.encoded_signer_key}")
        print(f"signer_ed25519 weight: {signer_ed25519.weight}")

    :param signer_key: The signer object
    :param weight: The weight of the key
    """

    def __init__(self, signer_key: SignerKey, weight: int) -> None:
        self.signer_key: SignerKey = signer_key
        self.weight: int = weight

    @classmethod
    def ed25519_public_key(cls, account_id: Union[str, bytes], weight: int) -> "Signer":
        """Create ED25519 PUBLIC KEY Signer from account id.

        :param account_id: account id (ex. ``"GDNA2V62PVEFBZ74CDJKTUHLY4Y7PL5UAV2MAM4VWF6USFE3SH2354AD"``)
        :param weight: The weight of the signer (0 to delete or 1-255)
        :return: ED25519 PUBLIC KEY Signer
        :raises:
            :exc:`Ed25519PublicKeyInvalidError <stellar_sdk.exceptions.Ed25519PublicKeyInvalidError>`: if ``account_id``
            is not a valid ed25519 public key.
        """
        signer_key = SignerKey.ed25519_public_key(account_id)
        return cls(signer_key, weight)

    @classmethod
    def pre_auth_tx(cls, pre_auth_tx_hash: Union[str, bytes], weight: int) -> "Signer":
        """Create Pre AUTH TX Signer from the sha256 hash of a transaction,
        click `here <https://developers.stellar.org/docs/glossary/multisig/#pre-authorized-transaction>`__ for more information.

        :param pre_auth_tx_hash: The sha256 hash of a transaction
            (ex. ``"TDNA2V62PVEFBZ74CDJKTUHLY4Y7PL5UAV2MAM4VWF6USFE3SH234BSS"`` or bytes)
        :param weight: The weight of the signer (0 to delete or 1-255)
        :return: Pre AUTH TX Signer
        """
        signer_key = SignerKey.pre_auth_tx(pre_auth_tx_hash)
        return cls(signer_key, weight)

    @classmethod
    def sha256_hash(cls, sha256_hash: Union[str, bytes], weight: int) -> "Signer":
        """Create SHA256 HASH Signer from a sha256 hash of a preimage,
        click `here <https://developers.stellar.org/docs/glossary/multisig/#hashx>`__ for more information.

        :param sha256_hash: a sha256 hash of a preimage
            (ex. ``"XDNA2V62PVEFBZ74CDJKTUHLY4Y7PL5UAV2MAM4VWF6USFE3SH235FXL"`` or bytes)
        :param weight: The weight of the signer (0 to delete or 1-255)
        :return: SHA256 HASH Signer
        """
        signer_key = SignerKey.sha256_hash(sha256_hash)
        return cls(signer_key, weight)

    def to_xdr_object(self) -> stellar_xdr.Signer:
        """Returns the xdr object for this Signer object.

        :return: XDR Signer object
        """
        return stellar_xdr.Signer(
            self.signer_key.to_xdr_object(), stellar_xdr.Uint32(self.weight)
        )

    @classmethod
    def from_xdr_object(cls, xdr_object: stellar_xdr.Signer) -> "Signer":
        """Create a :class:`Signer` from an XDR Signer object.

        :param xdr_object: The XDR Signer object.
        :return: A new :class:`Signer` object from the given XDR Signer object.
        """
        weight = xdr_object.weight.uint32
        signer_key = SignerKey.from_xdr_object(xdr_object.key)
        return cls(signer_key, weight)

    def __hash__(self):
        return hash((self.signer_key, self.weight))

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.signer_key == other.signer_key and self.weight == other.weight

    def __str__(self):
        return f"<Signer [signer_key={self.signer_key}, weight={self.weight}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64

from xdrlib3 import Packer, Unpacker

from .asset import Asset
from .int32 import Int32

__all__ = ["LiquidityPoolConstantProductParameters"]


class LiquidityPoolConstantProductParameters:
    """
    XDR Source Code::

        struct LiquidityPoolConstantProductParameters
        {
            Asset assetA; // assetA < assetB
            Asset assetB;
            int32 fee; // Fee is in basis points, so the actual rate is (fee/100)%
        };
    """

    def __init__(
        self,
        asset_a: Asset,
        asset_b: Asset,
        fee: Int32,
    ) -> None:
        self.asset_a = asset_a
        self.asset_b = asset_b
        self.fee = fee

    def pack(self, packer: Packer) -> None:
        self.asset_a.pack(packer)
        self.asset_b.pack(packer)
        self.fee.pack(packer)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> LiquidityPoolConstantProductParameters:
        asset_a = Asset.unpack(unpacker)
        asset_b = Asset.unpack(unpacker)
        fee = Int32.unpack(unpacker)
        return cls(
            asset_a=asset_a,
            asset_b=asset_b,
            fee=fee,
        )

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> LiquidityPoolConstantProductParameters:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> LiquidityPoolConstantProductParameters:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.asset_a,
                self.asset_b,
                self.fee,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.asset_a == other.asset_a
            and self.asset_b == other.asset_b
            and self.fee == other.fee
        )

    def __str__(self):
        out = [
            f"asset_a={self.asset_a}",
            f"asset_b={self.asset_b}",
            f"fee={self.fee}",
        ]
        return f"<LiquidityPoolConstantProductParameters [{', '.join(out)}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64

from xdrlib3 import Packer, Unpacker

from .base import String
from .hash import Hash
from .memo_type import MemoType
from .uint64 import Uint64

__all__ = ["Memo"]


class Memo:
    """
    XDR Source Code::

        union Memo switch (MemoType type)
        {
        case MEMO_NONE:
            void;
        case MEMO_TEXT:
            string text<28>;
        case MEMO_ID:
            uint64 id;
        case MEMO_HASH:
            Hash hash; // the hash of what to pull from the content server
        case MEMO_RETURN:
            Hash retHash; // the hash of the tx you are rejecting
        };
    """

    def __init__(
        self,
        type: MemoType,
        text: bytes = None,
        id: Uint64 = None,
        hash: Hash = None,
        ret_hash: Hash = None,
    ) -> None:
        self.type = type
        self.text = text
        self.id = id
        self.hash = hash
        self.ret_hash = ret_hash

    def pack(self, packer: Packer) -> None:
        self.type.pack(packer)
        if self.type == MemoType.MEMO_NONE:
            return
        if self.type == MemoType.MEMO_TEXT:
            if self.text is None:
                raise ValueError("text should not be None.")
            String(self.text, 28).pack(packer)
            return
        if self.type == MemoType.MEMO_ID:
            if self.id is None:
                raise ValueError("id should not be None.")
            self.id.pack(packer)
            return
        if self.type == MemoType.MEMO_HASH:
            if self.hash is None:
                raise ValueError("hash should not be None.")
            self.hash.pack(packer)
            return
        if self.type == MemoType.MEMO_RETURN:
            if self.ret_hash is None:
                raise ValueError("ret_hash should not be None.")
            self.ret_hash.pack(packer)
            return

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> Memo:
        type = MemoType.unpack(unpacker)
        if type == MemoType.MEMO_NONE:
            return cls(type=type)
        if type == MemoType.MEMO_TEXT:
            text = String.unpack(unpacker)
            return cls(type=type, text=text)
        if type == MemoType.MEMO_ID:
            id = Uint64.unpack(unpacker)
            return cls(type=type, id=id)
        if type == MemoType.MEMO_HASH:
            hash = Hash.unpack(unpacker)
            return cls(type=type, hash=hash)
        if type == MemoType.MEMO_RETURN:
            ret_hash = Hash.unpack(unpacker)
            return cls(type=type, ret_hash=ret_hash)
        return cls(type=type)

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> Memo:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> Memo:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.type,
                self.text,
                self.id,
                self.hash,
                self.ret_hash,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.type == other.type
            and self.text == other.text
            and self.id == other.id
            and self.hash == other.hash
            and self.ret_hash == other.ret_hash
        )

    def __str__(self):
        out = []
        out.append(f"type={self.type}")
        out.append(f"text={self.text}") if self.text is not None else None
        out.append(f"id={self.id}") if self.id is not None else None
        out.append(f"hash={self.hash}") if self.hash is not None else None
        out.append(f"ret_hash={self.ret_hash}") if self.ret_hash is not None else None
        return f"<Memo [{', '.join(out)}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64

from xdrlib3 import Packer, Unpacker

from .int64 import Int64
from .pool_id import PoolID
from .price import Price

__all__ = ["LiquidityPoolDepositOp"]


class LiquidityPoolDepositOp:
    """
    XDR Source Code::

        struct LiquidityPoolDepositOp
        {
            PoolID liquidityPoolID;
            int64 maxAmountA; // maximum amount of first asset to deposit
            int64 maxAmountB; // maximum amount of second asset to deposit
            Price minPrice;   // minimum depositA/depositB
            Price maxPrice;   // maximum depositA/depositB
        };
    """

    def __init__(
        self,
        liquidity_pool_id: PoolID,
        max_amount_a: Int64,
        max_amount_b: Int64,
        min_price: Price,
        max_price: Price,
    ) -> None:
        self.liquidity_pool_id = liquidity_pool_id
        self.max_amount_a = max_amount_a
        self.max_amount_b = max_amount_b
        self.min_price = min_price
        self.max_price = max_price

    def pack(self, packer: Packer) -> None:
        self.liquidity_pool_id.pack(packer)
        self.max_amount_a.pack(packer)
        self.max_amount_b.pack(packer)
        self.min_price.pack(packer)
        self.max_price.pack(packer)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> LiquidityPoolDepositOp:
        liquidity_pool_id = PoolID.unpack(unpacker)
        max_amount_a = Int64.unpack(unpacker)
        max_amount_b = Int64.unpack(unpacker)
        min_price = Price.unpack(unpacker)
        max_price = Price.unpack(unpacker)
        return cls(
            liquidity_pool_id=liquidity_pool_id,
            max_amount_a=max_amount_a,
            max_amount_b=max_amount_b,
            min_price=min_price,
            max_price=max_price,
        )

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> LiquidityPoolDepositOp:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> LiquidityPoolDepositOp:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.liquidity_pool_id,
                self.max_amount_a,
                self.max_amount_b,
                self.min_price,
                self.max_price,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.liquidity_pool_id == other.liquidity_pool_id
            and self.max_amount_a == other.max_amount_a
            and self.max_amount_b == other.max_amount_b
            and self.min_price == other.min_price
            and self.max_price == other.max_price
        )

    def __str__(self):
        out = [
            f"liquidity_pool_id={self.liquidity_pool_id}",
            f"max_amount_a={self.max_amount_a}",
            f"max_amount_b={self.max_amount_b}",
            f"min_price={self.min_price}",
            f"max_price={self.max_price}",
        ]
        return f"<LiquidityPoolDepositOp [{', '.join(out)}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64
from typing import List

from xdrlib3 import Packer, Unpacker

from .constants import *
from .hash import Hash

__all__ = ["TxDemandVector"]


class TxDemandVector:
    """
    XDR Source Code::

        typedef Hash TxDemandVector<TX_DEMAND_VECTOR_MAX_SIZE>;
    """

    def __init__(self, tx_demand_vector: List[Hash]) -> None:
        _expect_max_length = TX_DEMAND_VECTOR_MAX_SIZE
        if tx_demand_vector and len(tx_demand_vector) > _expect_max_length:
            raise ValueError(
                f"The maximum length of `tx_demand_vector` should be {_expect_max_length}, but got {len(tx_demand_vector)}."
            )
        self.tx_demand_vector = tx_demand_vector

    def pack(self, packer: Packer) -> None:
        packer.pack_uint(len(self.tx_demand_vector))
        for tx_demand_vector_item in self.tx_demand_vector:
            tx_demand_vector_item.pack(packer)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> TxDemandVector:
        length = unpacker.unpack_uint()
        tx_demand_vector = []
        for _ in range(length):
            tx_demand_vector.append(Hash.unpack(unpacker))
        return cls(tx_demand_vector)

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> TxDemandVector:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> TxDemandVector:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(self.tx_demand_vector)

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.tx_demand_vector == other.tx_demand_vector

    def __str__(self):
        return f"<TxDemandVector [tx_demand_vector={self.tx_demand_vector}]>"



from __future__ import annotations

import json
from typing import TYPE_CHECKING, Type

from . import Keypair
from . import xdr as stellar_xdr
from .account import Account
from .address import Address
from .base_soroban_server import (
    Durability,
    _assemble_transaction,
    _generate_unique_request_id,
)
from .client.requests_client import RequestsClient
from .exceptions import (
    AccountNotFoundException,
    PrepareTransactionException,
    SorobanRpcErrorResponse,
)
from .soroban_rpc import *

if TYPE_CHECKING:
    from .client.base_sync_client import BaseSyncClient
    from .transaction_envelope import TransactionEnvelope

__all__ = ["SorobanServer", "Durability"]

V = TypeVar("V")


class SorobanServer:
    """Server handles the network connection to a Soroban RPC instance and
    exposes an interface for requests to that instance.

    :param server_url: Soroban RPC server URL. (ex. ``https://rpc-futurenet.stellar.org:443/``)
    :param client: A client instance that will be used to make requests.
    """

    def __init__(
        self,
        server_url: str = "https://rpc-futurenet.stellar.org:443/",
        client: Optional[BaseSyncClient] = None,
    ) -> None:
        self.server_url: str = server_url

        if not client:
            client = RequestsClient()
        self._client: BaseSyncClient = client

    def get_health(self) -> GetHealthResponse:
        """General node health check.

        See `Soroban Documentation - getHealth <https://soroban.stellar.org/api/methods/getHealth>`_

        :return: A :class:`GetHealthResponse <stellar_sdk.soroban_rpc.GetHealthResponse>` object.
        :raises: :exc:`SorobanRpcErrorResponse <stellar_sdk.exceptions.SorobanRpcErrorResponse>` - If the Soroban-RPC instance returns an error response.
        """
        request: Request = Request(
            id=_generate_unique_request_id(),
            method="getHealth",
            params=None,
        )
        return self._post(request, GetHealthResponse)

    def get_events(
        self,
        start_ledger: int,
        filters: Sequence[EventFilter] = None,
        cursor: str = None,
        limit: int = None,
    ) -> GetEventsResponse:
        """Fetch a list of events that occurred in the ledger range.

        See `Soroban Documentation - getEvents <https://soroban.stellar.org/api/methods/getEvents>`_

        :param start_ledger: The first ledger to include in the results.
        :param filters: A list of filters to apply to the results.
        :param cursor: A cursor value for use in pagination.
        :param limit: The maximum number of records to return.
        :return: A :class:`GetEventsResponse <stellar_sdk.soroban_rpc.GetEventsResponse>` object.
        :raises: :exc:`SorobanRpcErrorResponse <stellar_sdk.exceptions.SorobanRpcErrorResponse>` - If the Soroban-RPC instance returns an error response.
        """
        pagination = PaginationOptions(cursor=cursor, limit=limit)
        data = GetEventsRequest(
            startLedger=str(start_ledger),
            filters=filters,
            pagination=pagination,
        )
        request: Request = Request[GetEventsRequest](
            id=_generate_unique_request_id(), method="getEvents", params=data
        )
        return self._post(request, GetEventsResponse)

    def get_network(self) -> GetNetworkResponse:
        """General info about the currently configured network.

        :return: A :class:`GetNetworkResponse <stellar_sdk.soroban_rpc.GetNetworkResponse>` object.
        :raises: :exc:`SorobanRpcErrorResponse <stellar_sdk.exceptions.SorobanRpcErrorResponse>` - If the Soroban-RPC instance returns an error response.
        """
        request: Request = Request(
            id=_generate_unique_request_id(),
            method="getNetwork",
            params=None,
        )
        return self._post(request, GetNetworkResponse)

    def get_latest_ledger(self) -> GetLatestLedgerResponse:
        """Fetches the latest ledger meta info from network which Soroban-RPC is connected to.

        :return: A :class:`GetLatestLedgerResponse <stellar_sdk.soroban_rpc.GetLatestLedgerResponse>` object.
        :raises: :exc:`SorobanRpcErrorResponse <stellar_sdk.exceptions.SorobanRpcErrorResponse>` - If the Soroban-RPC instance returns an error response.
        """
        request: Request = Request(
            id=_generate_unique_request_id(),
            method="getLatestLedger",
            params=None,
        )
        return self._post(request, GetLatestLedgerResponse)

    def get_ledger_entries(
        self, keys: List[stellar_xdr.LedgerKey]
    ) -> GetLedgerEntriesResponse:
        """For reading the current value of ledger entries directly.

        Allows you to directly inspect the current state of a contract, a contract's code,
        or any other ledger entry. This is a backup way to access your contract data
        which may not be available via events or simulateTransaction.

        See `Soroban Documentation - getLedgerEntries <https://soroban.stellar.org/api/methods/getLedgerEntries>`_

        :param keys: The ledger keys to fetch.
        :return: A :class:`GetLedgerEntriesResponse <stellar_sdk.soroban_rpc.GetLedgerEntryResponse>` object.
        :raises: :exc:`SorobanRpcErrorResponse <stellar_sdk.exceptions.SorobanRpcErrorResponse>` - If the Soroban-RPC instance returns an error response.
        """
        request = Request[GetLedgerEntriesRequest](
            id=_generate_unique_request_id(),
            method="getLedgerEntries",
            params=GetLedgerEntriesRequest(keys=[key.to_xdr() for key in keys]),
        )
        return self._post(request, GetLedgerEntriesResponse)

    def get_transaction(self, transaction_hash: str) -> GetTransactionResponse:
        """Fetch the specified transaction.

        See `Soroban Documentation - getTransaction <https://soroban.stellar.org/api/methods/getTransaction>`_

        :param transaction_hash: The hash of the transaction to fetch.
        :return: A :class:`GetTransactionResponse <stellar_sdk.soroban_rpc.GetTransactionResponse>` object.
        :raises: :exc:`SorobanRpcErrorResponse <stellar_sdk.exceptions.SorobanRpcErrorResponse>` - If the Soroban-RPC instance returns an error response.
        """
        request = Request[GetTransactionRequest](
            id=_generate_unique_request_id(),
            method="getTransaction",
            params=GetTransactionRequest(hash=transaction_hash),
        )
        return self._post(request, GetTransactionResponse)

    def simulate_transaction(
        self, transaction_envelope: TransactionEnvelope
    ) -> SimulateTransactionResponse:
        """Submit a trial contract invocation to get back return values, expected ledger footprint, and expected costs.

        See `Soroban Documentation - simulateTransaction <https://soroban.stellar.org/api/methods/simulateTransaction>`_

        :param transaction_envelope: The transaction to simulate. It should include exactly one operation,
            which must be one of :class:`RestoreFootprint <stellar_sdk.operation.RestoreFootprintOperation>`,
            :class:`InvokeHostFunction <stellar_sdk.operation.InvokeHostFunction>` or
            :class:`ExtendFootprintTTL <stellar_sdk.operation.RestoreFootprint>` operation.
            Any provided footprint will be ignored.
        :return: A :class:`SimulateTransactionResponse <stellar_sdk.soroban_rpc.SimulateTransactionResponse>` object
            contains the cost, footprint, result/auth requirements (if applicable), and error of the transaction.
        """
        xdr = (
            transaction_envelope
            if isinstance(transaction_envelope, str)
            else transaction_envelope.to_xdr()
        )
        request = Request[SimulateTransactionRequest](
            id=_generate_unique_request_id(),
            method="simulateTransaction",
            params=SimulateTransactionRequest(transaction=xdr),
        )
        return self._post(request, SimulateTransactionResponse)

    def send_transaction(
        self, transaction_envelope: Union[TransactionEnvelope, str]
    ) -> SendTransactionResponse:
        """Submit a real transaction to the Stellar network. This is the only way to make changes "on-chain".

        See `Soroban Documentation - sendTransaction <https://soroban.stellar.org/api/methods/sendTransaction>`_

        :param transaction_envelope: The transaction to send.
        :return: A :class:`SendTransactionResponse <stellar_sdk.soroban_rpc.SendTransactionResponse>` object.
        :raises: :exc:`SorobanRpcErrorResponse <stellar_sdk.exceptions.SorobanRpcErrorResponse>` - If the Soroban-RPC instance returns an error response.
        """
        xdr = (
            transaction_envelope
            if isinstance(transaction_envelope, str)
            else transaction_envelope.to_xdr()
        )
        request = Request[SendTransactionRequest](
            id=_generate_unique_request_id(),
            method="sendTransaction",
            params=SendTransactionRequest(transaction=xdr),
        )
        return self._post(request, SendTransactionResponse)

    def load_account(self, account_id: str) -> Account:
        """Load an account from the server, you can use the returned account
        object as source account for transactions.

        :param account_id: The account ID.
        :return: An :class:`Account <stellar_sdk.account.Account>` object.
        :raises: :exc:`AccountNotFoundException <stellar_sdk.exceptions.AccountNotFoundException>` - If the account is not found on the network.
        :raises: :exc:`SorobanRpcErrorResponse <stellar_sdk.exceptions.SorobanRpcErrorResponse>` - If the Soroban-RPC instance returns an error response.
        """
        account_id_xdr = Keypair.from_public_key(account_id).xdr_account_id()
        key = stellar_xdr.LedgerKey(
            stellar_xdr.LedgerEntryType.ACCOUNT,
            account=stellar_xdr.LedgerKeyAccount(account_id=account_id_xdr),
        )

        resp = self.get_ledger_entries([key])
        if not resp.entries:
            raise AccountNotFoundException(account_id)
        assert len(resp.entries) == 1
        data = stellar_xdr.LedgerEntryData.from_xdr(resp.entries[0].xdr)
        assert data.account is not None
        return Account(account_id, data.account.seq_num.sequence_number.int64)

    def get_contract_data(
        self,
        contract_id: str,
        key: stellar_xdr.SCVal,
        durability: Durability = Durability.PERSISTENT,
    ) -> Optional[LedgerEntryResult]:
        """Reads the current value of contract data ledger entries directly.

        :param contract_id: The contract ID containing the data to load. Encoded as Stellar Contract Address,
            for example: ``"CCJZ5DGASBWQXR5MPFCJXMBI333XE5U3FSJTNQU7RIKE3P5GN2K2WYD5"``
        :param key: The key of the contract data to load.
        :param durability: The "durability keyspace" that this ledger key belongs to, which is either
            :class:`Durability.TEMPORARY` or :class:`Durability.PERSISTENT`. Defaults to :class:`Durability.PERSISTENT`.
        :return: A :class:`LedgerEntryResult <stellar_sdk.soroban_rpc.LedgerEntryResult>` object contains the ledger entry result or ``None`` if not found.
        :raises: :exc:`SorobanRpcErrorResponse <stellar_sdk.exceptions.SorobanRpcErrorResponse>` - If the Soroban-RPC instance returns an error response.
        """
        sc_address = Address(contract_id).to_xdr_sc_address()
        xdr_durability = (
            stellar_xdr.ContractDataDurability.PERSISTENT
            if durability == Durability.PERSISTENT
            else stellar_xdr.ContractDataDurability.TEMPORARY
        )
        contract_key = stellar_xdr.LedgerKey(
            stellar_xdr.LedgerEntryType.CONTRACT_DATA,
            contract_data=stellar_xdr.LedgerKeyContractData(
                contract=sc_address,
                key=key,
                durability=xdr_durability,
            ),
        )
        resp = self.get_ledger_entries([contract_key])
        entries = resp.entries
        if not entries:
            return None
        return entries[0]

    def prepare_transaction(
        self,
        transaction_envelope: TransactionEnvelope,
        simulate_transaction_response: SimulateTransactionResponse = None,
    ) -> TransactionEnvelope:
        """Submit a trial contract invocation, first run a simulation of the contract
        invocation as defined on the incoming transaction, and apply the results to
        a new copy of the transaction which is then returned. Setting the ledger
        footprint and authorization, so the resulting transaction is ready for signing
        and sending.

        The returned transaction will also have an updated fee that is the sum of fee
        set on incoming transaction with the contract resource fees estimated from
        simulation. It is advisable to check the fee on returned transaction and validate
        or take appropriate measures for interaction with user to confirm it is acceptable.

        You can call the :meth:`simulate_transaction` method directly first if you
        want to inspect estimated fees for a given transaction in detail first if that is
        of importance.

        :param transaction_envelope: The transaction to prepare. It should include exactly one operation, which
            must be one of :py:class:`RestoreFootprint <stellar_sdk.operation.RestoreFootprint>`,
            :py:class:`ExtendFootprintTTL <stellar_sdk.operation.ExtendFootprintTTL>`,
            or :py:class:`InvokeHostFunction <stellar_sdk.operation.InvokeHostFunction>`. Any provided
            footprint will be ignored. You can use :meth:`stellar_sdk.Transaction.is_soroban_transaction` to check
            if a transaction is a Soroban transaction. Any provided footprint will be overwritten.
            However, if your operation has existing auth entries, they will be preferred over ALL auth
            entries from the simulation. In other words, if you include auth entries, you don't care
            about the auth returned from the simulation. Other fields (footprint, etc.) will be filled
            as normal.
        :param simulate_transaction_response: The response of the simulation of the transaction,
            typically you don't need to pass this parameter, it will be automatically called if you don't pass it.
        :return: A copy of the :class:`TransactionEnvelope <stellar_sdk.transaction_envelope.TransactionEnvelope>`,
            with the expected authorizations (in the case of invocation) and ledger footprint added.
            The transaction fee will also automatically be padded with the contract's minimum resource fees
            discovered from the simulation.
        """
        if not simulate_transaction_response:
            simulate_transaction_response = self.simulate_transaction(
                transaction_envelope
            )
        if simulate_transaction_response.error:
            raise PrepareTransactionException(
                "Simulation transaction failed, the response contains error information.",
                simulate_transaction_response,
            )
        te = _assemble_transaction(transaction_envelope, simulate_transaction_response)
        return te

    def close(self) -> None:
        """Close underlying connector, and release all acquired resources."""
        self._client.close()

    def _post(self, request_body: Request, response_body_type: Type[V]) -> V:
        json_data = request_body.model_dump_json(by_alias=True)
        data = self._client.post(
            self.server_url,
            json_data=json.loads(json_data),
        )
        response = Response[response_body_type].model_validate(data.json())  # type: ignore[valid-type]
        if response.error:
            raise SorobanRpcErrorResponse(
                response.error.code, response.error.message, response.error.data
            )
        assert response.result is not None
        return response.result

    def __enter__(self) -> "SorobanServer":
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def __str__(self):
        return f"<SorobanServer [server_url={self.server_url}, client={self._client}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64

from xdrlib3 import Packer, Unpacker

from .auth_cert import AuthCert
from .base import Integer, String
from .hash import Hash
from .node_id import NodeID
from .uint32 import Uint32
from .uint256 import Uint256

__all__ = ["Hello"]


class Hello:
    """
    XDR Source Code::

        struct Hello
        {
            uint32 ledgerVersion;
            uint32 overlayVersion;
            uint32 overlayMinVersion;
            Hash networkID;
            string versionStr<100>;
            int listeningPort;
            NodeID peerID;
            AuthCert cert;
            uint256 nonce;
        };
    """

    def __init__(
        self,
        ledger_version: Uint32,
        overlay_version: Uint32,
        overlay_min_version: Uint32,
        network_id: Hash,
        version_str: bytes,
        listening_port: int,
        peer_id: NodeID,
        cert: AuthCert,
        nonce: Uint256,
    ) -> None:
        self.ledger_version = ledger_version
        self.overlay_version = overlay_version
        self.overlay_min_version = overlay_min_version
        self.network_id = network_id
        self.version_str = version_str
        self.listening_port = listening_port
        self.peer_id = peer_id
        self.cert = cert
        self.nonce = nonce

    def pack(self, packer: Packer) -> None:
        self.ledger_version.pack(packer)
        self.overlay_version.pack(packer)
        self.overlay_min_version.pack(packer)
        self.network_id.pack(packer)
        String(self.version_str, 100).pack(packer)
        Integer(self.listening_port).pack(packer)
        self.peer_id.pack(packer)
        self.cert.pack(packer)
        self.nonce.pack(packer)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> Hello:
        ledger_version = Uint32.unpack(unpacker)
        overlay_version = Uint32.unpack(unpacker)
        overlay_min_version = Uint32.unpack(unpacker)
        network_id = Hash.unpack(unpacker)
        version_str = String.unpack(unpacker)
        listening_port = Integer.unpack(unpacker)
        peer_id = NodeID.unpack(unpacker)
        cert = AuthCert.unpack(unpacker)
        nonce = Uint256.unpack(unpacker)
        return cls(
            ledger_version=ledger_version,
            overlay_version=overlay_version,
            overlay_min_version=overlay_min_version,
            network_id=network_id,
            version_str=version_str,
            listening_port=listening_port,
            peer_id=peer_id,
            cert=cert,
            nonce=nonce,
        )

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> Hello:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> Hello:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.ledger_version,
                self.overlay_version,
                self.overlay_min_version,
                self.network_id,
                self.version_str,
                self.listening_port,
                self.peer_id,
                self.cert,
                self.nonce,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.ledger_version == other.ledger_version
            and self.overlay_version == other.overlay_version
            and self.overlay_min_version == other.overlay_min_version
            and self.network_id == other.network_id
            and self.version_str == other.version_str
            and self.listening_port == other.listening_port
            and self.peer_id == other.peer_id
            and self.cert == other.cert
            and self.nonce == other.nonce
        )

    def __str__(self):
        out = [
            f"ledger_version={self.ledger_version}",
            f"overlay_version={self.overlay_version}",
            f"overlay_min_version={self.overlay_min_version}",
            f"network_id={self.network_id}",
            f"version_str={self.version_str}",
            f"listening_port={self.listening_port}",
            f"peer_id={self.peer_id}",
            f"cert={self.cert}",
            f"nonce={self.nonce}",
        ]
        return f"<Hello [{', '.join(out)}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64

from xdrlib3 import Packer, Unpacker

from .account_id import AccountID
from .asset_code import AssetCode
from .uint32 import Uint32

__all__ = ["AllowTrustOp"]


class AllowTrustOp:
    """
    XDR Source Code::

        struct AllowTrustOp
        {
            AccountID trustor;
            AssetCode asset;

            // One of 0, AUTHORIZED_FLAG, or AUTHORIZED_TO_MAINTAIN_LIABILITIES_FLAG
            uint32 authorize;
        };
    """

    def __init__(
        self,
        trustor: AccountID,
        asset: AssetCode,
        authorize: Uint32,
    ) -> None:
        self.trustor = trustor
        self.asset = asset
        self.authorize = authorize

    def pack(self, packer: Packer) -> None:
        self.trustor.pack(packer)
        self.asset.pack(packer)
        self.authorize.pack(packer)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> AllowTrustOp:
        trustor = AccountID.unpack(unpacker)
        asset = AssetCode.unpack(unpacker)
        authorize = Uint32.unpack(unpacker)
        return cls(
            trustor=trustor,
            asset=asset,
            authorize=authorize,
        )

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> AllowTrustOp:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> AllowTrustOp:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.trustor,
                self.asset,
                self.authorize,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.trustor == other.trustor
            and self.asset == other.asset
            and self.authorize == other.authorize
        )

    def __str__(self):
        out = [
            f"trustor={self.trustor}",
            f"asset={self.asset}",
            f"authorize={self.authorize}",
        ]
        return f"<AllowTrustOp [{', '.join(out)}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64

from xdrlib3 import Packer, Unpacker

from .extension_point import ExtensionPoint
from .int64 import Int64

__all__ = ["ContractCostParamEntry"]


class ContractCostParamEntry:
    """
    XDR Source Code::

        struct ContractCostParamEntry {
            // use `ext` to add more terms (e.g. higher order polynomials) in the future
            ExtensionPoint ext;

            int64 constTerm;
            int64 linearTerm;
        };
    """

    def __init__(
        self,
        ext: ExtensionPoint,
        const_term: Int64,
        linear_term: Int64,
    ) -> None:
        self.ext = ext
        self.const_term = const_term
        self.linear_term = linear_term

    def pack(self, packer: Packer) -> None:
        self.ext.pack(packer)
        self.const_term.pack(packer)
        self.linear_term.pack(packer)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> ContractCostParamEntry:
        ext = ExtensionPoint.unpack(unpacker)
        const_term = Int64.unpack(unpacker)
        linear_term = Int64.unpack(unpacker)
        return cls(
            ext=ext,
            const_term=const_term,
            linear_term=linear_term,
        )

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> ContractCostParamEntry:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> ContractCostParamEntry:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.ext,
                self.const_term,
                self.linear_term,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.ext == other.ext
            and self.const_term == other.const_term
            and self.linear_term == other.linear_term
        )

    def __str__(self):
        out = [
            f"ext={self.ext}",
            f"const_term={self.const_term}",
            f"linear_term={self.linear_term}",
        ]
        return f"<ContractCostParamEntry [{', '.join(out)}]>"
